import logging
import os
import time
from pathlib import Path

import jax
import numpy as np
from tqdm import trange

from apax.config import parse_config
from apax.data.input_pipeline import OTFInMemoryDataset
from apax.train.callbacks import initialize_callbacks
from apax.train.checkpoints import restore_single_parameters
from apax.train.metrics import initialize_metrics
from apax.train.run import initialize_loss_fn, setup_logging
from apax.train.trainer import make_step_fns
from apax.utils.data import load_data, split_atoms
from apax.utils.random import seed_py_np_tf

log = logging.getLogger(__name__)


def get_test_idxs(atoms_list, used_idxs, n_test=-1):
    idxs = np.arange(len(atoms_list))
    test_idxs = np.setdiff1d(idxs, used_idxs)
    np.random.shuffle(test_idxs)
    if n_test != -1:
        test_idxs = test_idxs[:n_test]

    return test_idxs


def load_test_data(
    config, model_version_path, eval_path, n_test=-1
):  # TODO double code run.py in progress
    """
    Load test data for evaluation.

    Parameters
    ----------
    config : object
        Configuration object.
    model_version_path : str
        Path to the model version.
    eval_path : str
        Path to evaluation directory.
    n_test : int, default = -1
        Number of test samples to load, by default -1 (load all).

    Returns
    -------
    atoms_list
        List of ase.Atoms containing the test data.
    """

    log.info("Running Input Pipeline")
    os.makedirs(eval_path, exist_ok=True)

    if config.data.test_data_path is not None:
        log.info(f"Read test data file {config.data.test_data_path}")
        atoms_list = load_data(config.data.test_data_path)
        atoms_list = atoms_list[:n_test]

    elif config.data.data_path is not None:
        log.info(f"Read data file {config.data.data_path}")
        atoms_list = load_data(config.data.data_path)

        idxs_dict = np.load(model_version_path / "train_val_idxs.npz")

        used_idxs = idxs_dict["train_idxs"]
        np.append(used_idxs, idxs_dict["val_idxs"])

        test_idxs = get_test_idxs(atoms_list, used_idxs, n_test)

        np.savez(
            os.path.join(eval_path, "test_idxs"),
            test_idxs=test_idxs,
        )

        atoms_list, _ = split_atoms(atoms_list, test_idxs)

    else:
        raise ValueError("input data path/paths not defined")

    return atoms_list


def predict(model, params, Metrics, loss_fn, test_ds, callbacks, is_ensemble=False):
    """
    Perform predictions on the test dataset.

    Parameters
    ----------
    model :
        Trained model.
    params :
        Model parameters.
    Metrics :
        Collection of metrics.
    loss_fn :
        Loss function.
    test_ds :
        Test dataset.
    callbacks :
        Callback functions.
    is_ensemble : bool, default = False
        Whether the model is an ensemble.
    """

    callbacks.on_train_begin()
    _, test_step_fn = make_step_fns(
        loss_fn, Metrics, model=model, is_ensemble=is_ensemble
    )

    batch_test_ds = test_ds.batch()

    test_metrics = Metrics.empty()

    batch_pbar = trange(
        0, test_ds.n_data, desc="Structure", ncols=100, disable=False, leave=True
    )
    for batch_idx in range(test_ds.n_data):
        batch = next(batch_test_ds)
        batch_start_time = time.time()

        batch_loss, test_metrics = test_step_fn(params, batch, test_metrics)
        batch_metrics = {"test_loss": float(batch_loss)}
        batch_metrics.update(
            {f"test_{key}": float(val) for key, val in test_metrics.compute().items()}
        )
        batch_end_time = time.time()
        batch_metrics.update({"time": batch_end_time - batch_start_time})

        callbacks.on_test_batch_end(batch=batch_idx, logs=batch_metrics)

        batch_pbar.set_postfix(test_loss=batch_metrics["test_loss"])
        batch_pbar.update()
    batch_pbar.close()
    callbacks.on_train_end()


def eval_model(config_path, n_test=-1, log_file="eval.log", log_level="error"):
    """
    Evaluate the model using the provided configuration.

    Parameters
    ----------
    config_path : str
        Path to the configuration file.
    n_test : int, default = -1
        Number of test samples to evaluate, by default -1 (evaluate all).
    log_file : str, default = "eval.log"
        Path to the log file.
    log_level : str, default = "error"
        Logging level.
    """

    setup_logging(log_file, log_level)
    log.info("Starting model evaluation")
    config = parse_config(config_path)

    seed_py_np_tf(config.seed)

    model_version_path = Path(config.data.directory) / config.data.experiment
    eval_path = model_version_path / "eval"

    callbacks = initialize_callbacks(config, eval_path)
    loss_fn = initialize_loss_fn(config.loss)
    Metrics = initialize_metrics(config.metrics)

    atoms_list = load_test_data(config, model_version_path, eval_path, n_test)
    test_ds = OTFInMemoryDataset(
        atoms_list,
        config.model.basis.r_max,
        1,
        config.data.valid_batch_size,
        pos_unit=config.data.pos_unit,
        energy_unit=config.data.energy_unit,
    )

    _, init_box = test_ds.init_input()

    if config.model.ensemble and config.model.ensemble.kind == "full":
        n_full_models = config.model.ensemble.n_members
    else:
        n_full_models = 1

    Builder = config.model.get_builder()
    builder = Builder(config.model.model_dump(), n_species=119)
    model = builder.build_energy_derivative_model(
        apply_mask=True,
        init_box=init_box,
    )

    model = jax.vmap(model.apply, in_axes=(None, 0, 0, 0, 0, 0))
    config, params = restore_single_parameters(model_version_path)

    predict(
        model,
        params,
        Metrics,
        loss_fn,
        test_ds,
        callbacks,
        is_ensemble=n_full_models > 1,
    )



import jax.numpy as jnp
import numpy as np
from jax import vmap

from apax.utils.jax_md_reduced import partition, space


def canonicalize_neighbors(neighbor):
    return neighbor.idx if isinstance(neighbor, partition.NeighborList) else neighbor


def disp_fn(ri, rj, perturbation, box):
    dR = space.pairwise_displacement(ri, rj)
    dR = space.transform(box, dR)

    if perturbation is not None:
        dR = dR + space.raw_transform(perturbation, dR)
        # https://github.com/mir-group/nequip/blob/c56f48fcc9b4018a84e1ed28f762fadd5bc763f1/nequip/nn/_grad_output.py#L267
        # https://github.com/sirmarcel/glp/blob/main/glp/calculators/utils.py
        # other codes do R = R + strain, not dR
        # can be implemented for efficiency
    return dR


def get_disp_fn(displacement):
    def disp_fn(ri, rj, perturbation, box):
        return displacement(ri, rj, perturbation, box=box)

    return disp_fn


def make_distance_fn(init_box, inference_disp_fn=None):
    """Model which post processes the output of an atomistic model and
    adds empirical energy terms.
    """

    if np.all(init_box < 1e-6):
        # gas phase training and predicting
        displacement_fn = space.free()[0]
        displacement = space.map_bond(displacement_fn)
    elif inference_disp_fn is None:
        # for training on periodic systems
        displacement = vmap(disp_fn, (0, 0, None, None), 0)
    else:
        mappable_displacement_fn = get_disp_fn(inference_disp_fn)
        displacement = vmap(mappable_displacement_fn, (0, 0, None, None), 0)

    def compute_distances(R, neighbor, box, offsets, perturbation=None):
        # Distances
        idx = canonicalize_neighbors(neighbor)
        idx_i, idx_j = idx[0], idx[1]

        # R shape n_atoms x 3
        R = R.astype(jnp.float64)
        Ri = R[idx_i]
        Rj = R[idx_j]

        # dr_vec shape: neighbors x 3
        if np.all(init_box < 1e-6):
            # reverse conventnion to match TF
            # distance vector for gas phase training and predicting
            dr_vec = displacement(Rj, Ri)
        else:
            # distance vector for training on periodic systems
            dr_vec = displacement(Rj, Ri, perturbation, box)
            dr_vec += offsets
        return dr_vec, idx

    return compute_distances



import einops
import numpy as np


class KernelMatrix:
    """
    Matrix representation of a kernel defined by a feature map g
    K_{ij} = \\sum_{k} g_{ik} g_{jk}
    """

    def __init__(self, g: np.ndarray, n_train: int):
        self.num_columns = g.shape[0]
        self.g = g
        self.diagonal = einops.einsum(g, g, "s feature, s feature -> s")
        self.n_train = n_train

    def compute_column(self, idx: int) -> np.ndarray:
        return einops.einsum(self.g, self.g[idx, :], "s feature, feature -> s")

    def score(self, idx: int) -> np.ndarray:
        """Computes the distance of sample i from all other samples j as
        K_{ii} + K_{jj} - 2 K_{ij}
        """
        return self.diagonal[idx] + self.diagonal - 2 * self.compute_column(idx)



import jax
from jax import tree_util


@jax.jit
def tree_ema(tree1, tree2, alpha):
    """Exponential moving average of two pytrees."""
    ema = tree_util.tree_map(lambda a, b: alpha * a + (1 - alpha) * b, tree1, tree2)
    return ema


class EMAParameters:
    """Handler for tracking an exponential moving average of model parameters.
    The EMA parameters are used in the valitaion loop.

    Parameters
    ----------
    ema_start : int, default = 1
        Epoch at which to start averaging models.
    alpha : float, default = 0.9
        How much of the new model to use. 1.0 would mean no averaging, 0.0 no updates.
    """

    def __init__(self, ema_start: int, alpha: float = 0.9) -> None:
        self.alpha = alpha
        self.ema_start = ema_start
        self.ema_params = None

    def update(self, opt_params, epoch):
        if epoch > self.ema_start:
            self.ema_params = tree_ema(opt_params, self.ema_params, self.alpha)
        else:
            self.ema_params = opt_params



import logging
from functools import partial

import jax.numpy as jnp
from clu import metrics

from apax.utils.math import normed_dotp

log = logging.getLogger(__name__)


class Averagefp64(metrics.Average):
    @classmethod
    def empty(cls) -> metrics.Metric:
        return cls(total=jnp.array(0, jnp.float64), count=jnp.array(0, jnp.int64))


class RootAverage(Averagefp64):
    """
    Modifies the `compute` method of `metrics.Average` to obtain the root of the average.
    Meant to be used with `mse_fn`.
    """

    def compute(self) -> jnp.array:
        return jnp.sqrt(self.total / self.count)


def mae_fn(label: dict[jnp.array], prediction: dict[jnp.array], key: str) -> jnp.array:
    """
    Computes the Mean Absolute Error of two arrays.
    """
    return jnp.mean(jnp.abs(label[key] - prediction[key]))


def mse_fn(label: dict[jnp.array], prediction: dict[jnp.array], key: str) -> jnp.array:
    """
    Computes the Mean Squared Error of two arrays.
    """
    return jnp.mean((label[key] - prediction[key]) ** 2)


def cosine_sim(
    label: dict[jnp.array], prediction: dict[jnp.array], key: str
) -> jnp.array:
    """
    Computes the cosine similarity of two arrays.
    """
    dotp = normed_dotp(label[key], prediction[key])
    F_angle_loss = jnp.mean(1.0 - dotp)
    return F_angle_loss


def make_single_metric(key: str, reduction: str) -> metrics.Average:
    """
    Builds a single `clu` metric where the key picks out the quantity from
    the model predictions dict.
    Metric functions (like `mae_fn`) are curried with the `key`.
    """
    reduction_fns = {
        "mae": mae_fn,
        "mse": mse_fn,
        "rmse": mse_fn,
        "cosine_sim": cosine_sim,
    }
    if reduction == "rmse":
        metric = RootAverage
    else:
        metric = Averagefp64

    reduction_fn = reduction_fns[reduction]
    reduction_fn = partial(reduction_fn, key=key)

    return metric.from_fun(reduction_fn)


def initialize_metrics(metrics_list) -> metrics.Collection:
    """
    Builds a `clu` metrics `Collection` by looping over all `keys` and `reductions`.
    the metrics are named according to `key_reduction`.
    See `make_single_metric` for details on the individual metrics.
    """
    log.info("Initializing Metrics")
    keys = []
    reductions = []
    for metric in metrics_list:
        for reduction in metric.reductions:
            keys.append(metric.name)
            reductions.append(reduction)

    metric_dict = {}
    for key, reduction in zip(keys, reductions):
        metric = make_single_metric(key, reduction)
        metric_identifier = f"{key}_{reduction}"
        metric_dict[metric_identifier] = metric

    metrics_collection = metrics.Collection.create(**metric_dict)

    return metrics_collection



from typing import Tuple

import numpy as np
import uncertainty_toolbox as uct
from ase import Atoms

from apax.md import ASECalculator


def compute_calibration_factors(
    calc: ASECalculator,
    atoms_list: list[Atoms],
    batch_size: int = 32,
    criterion: str = "ma_cal",
    shared_factor=False,
    optimizer_bounds: Tuple[float, float] = (1e-2, 1e2),
) -> Tuple[float, float]:
    """Computes global recalibration factors. These can be supplied to the ASEcalculator as part of the GlobalCalibration transformation.

    Parameters
    ----------
    calc: ASECalculator
        Model to be calibrated
    atoms_list: list[Atoms]
        Calibration dataset
    batch_size: int, default = 32
        Processing batch size. Choose the largest allowed by your VRAM.
    criterion: str, default = "ma_cal
        Calibration criterion. See uncertainty_toolbox for more details.
    shared_factor: bool, default = False
        Whether or not to calibrate energies and forces separately.
    optimizer_bounds: Tuple[float, float], default = (1e-2, 1e2)
        Search value bounds.

    Returns
    -------
    e_factor: float
        Global energy calibration factor.
    f_factor: float
        Global forces calibration factor.
    """

    num_atoms = np.array([len(a) for a in atoms_list])
    Etrue = np.array([a.get_potential_energy() for a in atoms_list]) / num_atoms
    Ftrue = np.reshape([a.get_forces() for a in atoms_list], (-1,))
    new_atoms = calc.batch_eval(atoms_list, batch_size=batch_size)

    Epred = np.array([a.get_potential_energy() for a in new_atoms]) / num_atoms
    Fpred = np.reshape([a.get_forces() for a in new_atoms], (-1,))

    Estd = np.array([a.calc.results["energy_uncertainty"] for a in new_atoms]) / num_atoms
    Fstd = np.reshape([a.calc.results["forces_uncertainty"] for a in new_atoms], (-1,))

    e_factor = uct.optimize_recalibration_ratio(
        Epred, Estd, Etrue, criterion, optimizer_bounds=optimizer_bounds
    )

    if shared_factor:
        f_factor = e_factor
    else:
        f_factor = uct.optimize_recalibration_ratio(
            Fpred, Fstd, Ftrue, criterion, optimizer_bounds=optimizer_bounds
        )

    return e_factor, f_factor



import logging
import multiprocessing
import time
import uuid
from collections import deque
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from pathlib import Path
from queue import Queue
from random import shuffle
from threading import Event
from typing import Dict, Iterator, Optional

import jax
import jax.numpy as jnp
import numpy as np
import tensorflow as tf
from jax import tree_util

from apax.data.preprocessing import compute_nl, prefetch_to_single_device
from apax.utils.convert import (
    atoms_to_inputs,
    atoms_to_labels,
    transpose_dict_of_lists,
    unit_dict,
)

log = logging.getLogger(__name__)


def pad_nl(idx, offsets, max_neighbors):
    """
    Pad the neighbor list arrays to the maximal number of neighbors occurring.

    Parameters
    ----------
    idx : np.ndarray
        Neighbor indices array.
    offsets : np.ndarray
        Offset array.
    max_neighbors : int
        Maximum number of neighbors.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Tuple containing padded neighbor indices array and offsets array.
    """
    zeros_to_add = max_neighbors - idx.shape[1]
    idx = np.pad(idx, ((0, 0), (0, zeros_to_add)), "constant").astype(np.int16)
    offsets = np.pad(offsets, ((0, zeros_to_add), (0, 0)), "constant")
    return idx, offsets


def find_largest_system(inputs, r_max) -> tuple[int]:
    """
    Finds the maximal number of atoms and neighbors.

    Parameters
    ----------
    inputs : dict
        Dictionary containing input data.
    r_max : float
        Maximum interaction radius.

    Returns
    -------
    Tuple[int]
        Tuple containing the maximum number of atoms and neighbors.
    """
    positions, boxes = inputs["positions"], inputs["box"]
    max_atoms = np.max(inputs["n_atoms"])

    max_nbrs = 0
    for pos, box in zip(positions, boxes):
        neighbor_idxs, _ = compute_nl(pos, box, r_max)
        n_neighbors = neighbor_idxs.shape[1]
        max_nbrs = max(max_nbrs, n_neighbors)

    return max_atoms, max_nbrs


class InMemoryDataset:
    """Baseclass for all datasets which store data in memory."""

    def __init__(
        self,
        atoms_list,
        cutoff,
        bs,
        n_epochs,
        n_jit_steps=1,
        pos_unit: str = "Ang",
        energy_unit: str = "eV",
        additional_properties: list[tuple] = [],
        pre_shuffle=False,
        shuffle_buffer_size=1000,
        ignore_labels=False,
        cache_path=".",
    ) -> None:
        self.n_epochs = n_epochs
        self.cutoff = cutoff
        self.n_jit_steps = n_jit_steps
        self.buffer_size = shuffle_buffer_size
        self.n_data = len(atoms_list)
        self.batch_size = self.validate_batch_size(bs)
        self.pos_unit = pos_unit
        self.additional_properties = additional_properties

        if pre_shuffle:
            shuffle(atoms_list)
        self.sample_atoms = atoms_list[0]
        self.inputs = atoms_to_inputs(atoms_list, pos_unit)

        max_atoms, max_nbrs = find_largest_system(self.inputs, self.cutoff)
        self.max_atoms = max_atoms
        self.max_nbrs = max_nbrs
        if atoms_list[0].calc and not ignore_labels:
            self.labels = atoms_to_labels(
                atoms_list, pos_unit, energy_unit, additional_properties
            )
        else:
            self.labels = None

        self.count = 0
        self.buffer = deque()
        self.file = Path(cache_path) / str(uuid.uuid4())

        self.enqueue(min(self.buffer_size, self.n_data))

    def steps_per_epoch(self) -> int:
        """Returns the number of steps per epoch dependent on the number of data and the
        batch size. Steps per epoch are calculated in a way that all epochs have the same
        number of steps, and all batches have the same length. To do so, some training
        data are dropped in each epoch.
        """
        return self.n_data // self.batch_size // self.n_jit_steps

    def validate_batch_size(self, batch_size: int) -> int:
        if batch_size > self.n_data:
            msg = (
                f"requested batch size {batch_size} is larger than the number of data"
                f" points {self.n_data}. Setting batch size = {self.n_data}"
            )
            log.warning(msg)
            batch_size = self.n_data
        return batch_size

    def prepare_data(self, i):
        inputs = {k: v[i] for k, v in self.inputs.items()}
        idx, offsets = compute_nl(inputs["positions"], inputs["box"], self.cutoff)
        inputs["idx"], inputs["offsets"] = pad_nl(idx, offsets, self.max_nbrs)

        zeros_to_add = self.max_atoms - inputs["numbers"].shape[0]
        inputs["positions"] = np.pad(
            inputs["positions"], ((0, zeros_to_add), (0, 0)), "constant"
        )
        inputs["numbers"] = np.pad(
            inputs["numbers"], (0, zeros_to_add), "constant"
        ).astype(np.int16)

        if not self.labels:
            return inputs

        labels = {k: v[i] for k, v in self.labels.items()}
        if "forces" in labels:
            labels["forces"] = np.pad(
                labels["forces"], ((0, zeros_to_add), (0, 0)), "constant"
            )

        for prop in self.additional_properties:
            name, shape = prop
            if shape[0] == "natoms":
                pad_shape = [(0, zeros_to_add)] + [(0, 0)] * (len(shape) - 1)
                labels[name] = np.pad(labels[name], pad_shape, "constant")

        inputs = {k: tf.constant(v) for k, v in inputs.items()}
        labels = {k: tf.constant(v) for k, v in labels.items()}

        return (inputs, labels)

    def enqueue(self, num_elements):
        for _ in range(num_elements):
            data = self.prepare_data(self.count)
            self.buffer.append(data)
            self.count += 1

    def make_signature(self) -> tf.TensorSpec:
        input_signature = {}
        input_signature["n_atoms"] = tf.TensorSpec((), dtype=tf.int16, name="n_atoms")
        input_signature["numbers"] = tf.TensorSpec(
            (self.max_atoms,), dtype=tf.int16, name="numbers"
        )
        input_signature["positions"] = tf.TensorSpec(
            (self.max_atoms, 3), dtype=tf.float64, name="positions"
        )
        input_signature["box"] = tf.TensorSpec((3, 3), dtype=tf.float64, name="box")
        input_signature["idx"] = tf.TensorSpec(
            (2, self.max_nbrs), dtype=tf.int16, name="idx"
        )
        input_signature["offsets"] = tf.TensorSpec(
            (self.max_nbrs, 3), dtype=tf.float64, name="offsets"
        )

        if not self.labels:
            return input_signature

        label_signature = {}
        if "energy" in self.labels.keys():
            label_signature["energy"] = tf.TensorSpec((), dtype=tf.float64, name="energy")
        if "forces" in self.labels.keys():
            label_signature["forces"] = tf.TensorSpec(
                (self.max_atoms, 3), dtype=tf.float64, name="forces"
            )
        if "stress" in self.labels.keys():
            label_signature["stress"] = tf.TensorSpec(
                (3, 3), dtype=tf.float64, name="stress"
            )

        for prop in self.additional_properties:
            name, shape = prop
            if shape[0] == "natoms":
                shape[0] = self.max_atoms

            sig = tf.TensorSpec(tuple(shape), dtype=tf.float64, name=name)
            label_signature[name] = sig
        signature = (input_signature, label_signature)
        return signature

    def init_input(self) -> Dict[str, np.ndarray]:
        """Returns first batch of inputs and labels to init the model."""
        positions = self.sample_atoms.positions * unit_dict[self.pos_unit]
        box = self.sample_atoms.cell.array * unit_dict[self.pos_unit]
        # For an input sample, it does not matter whether pos is fractional or cartesian
        idx, offsets = compute_nl(positions, box, self.cutoff)
        inputs = (
            positions,
            self.sample_atoms.numbers,
            idx,
            box,
            offsets,
        )

        inputs = tree_util.tree_map(lambda x: jnp.array(x), inputs)
        return inputs, np.array(box)

    def __iter__(self):
        raise NotImplementedError

    def shuffle_and_batch(self):
        raise NotImplementedError

    def batch(self) -> Iterator[jax.Array]:
        raise NotImplementedError

    def cleanup(self):
        pass


class CachedInMemoryDataset(InMemoryDataset):
    """Dataset which pads everything (atoms, neighbors)
    to the largest system in the dataset.
    The NL is computed on the fly during the first epoch and stored to disk using
    tf.data's cache.
    Most performant option for datasets with samples of very similar size.
    """

    def __iter__(self):
        while self.count < self.n_data or len(self.buffer) > 0:
            yield self.buffer.popleft()

            space = self.buffer_size - len(self.buffer)
            if self.count + space > self.n_data:
                space = self.n_data - self.count
            self.enqueue(space)

    def shuffle_and_batch(self, sharding=None):
        """Shuffles and batches the inputs/labels. This function prepares the
        inputs and labels for the whole training and prefetches the data.

        Returns
        -------
        ds :
            Iterator that returns inputs and labels of one batch in each step.
        """
        ds = (
            tf.data.Dataset.from_generator(
                lambda: self, output_signature=self.make_signature()
            )
            .cache(self.file.as_posix())
            .repeat(self.n_epochs)
        )

        ds = ds.shuffle(
            buffer_size=self.buffer_size, reshuffle_each_iteration=True
        ).batch(batch_size=self.batch_size)
        if self.n_jit_steps > 1:
            ds = ds.batch(batch_size=self.n_jit_steps)
        ds = prefetch_to_single_device(
            ds.as_numpy_iterator(), 2, sharding, n_step_jit=self.n_jit_steps > 1
        )
        return ds

    def batch(self, sharding=None) -> Iterator[jax.Array]:
        ds = (
            tf.data.Dataset.from_generator(
                lambda: self, output_signature=self.make_signature()
            )
            .cache(self.file.as_posix())
            .repeat(self.n_epochs)
        )
        ds = ds.batch(batch_size=self.batch_size)
        ds = prefetch_to_single_device(
            ds.as_numpy_iterator(), 2, sharding, n_step_jit=self.n_jit_steps > 1
        )
        return ds

    def cleanup(self):
        for p in self.file.parent.glob(f"{self.file.name}.data*"):
            p.unlink()

        index_file = self.file.parent / f"{self.file.name}.index"
        index_file.unlink()


class OTFInMemoryDataset(InMemoryDataset):
    """Dataset which pads everything (atoms, neighbors)
    to the largest system in the dataset.
    The NL is computed on the fly and fed into a tf.data generator.
    Mostly for internal purposes.
    """

    def __iter__(self):
        outer_count = 0
        max_iter = self.n_data * self.n_epochs
        while outer_count < max_iter:
            yield self.buffer.popleft()

            space = self.buffer_size - len(self.buffer)
            if self.count + space > self.n_data:
                space = self.n_data - self.count

            if self.count >= self.n_data:
                self.count = 0
            self.enqueue(space)
            outer_count += 1

    def shuffle_and_batch(self, sharding=None):
        """Shuffles and batches the inputs/labels. This function prepares the
        inputs and labels for the whole training and prefetches the data.

        Returns
        -------
        ds :
            Iterator that returns inputs and labels of one batch in each step.
        """
        ds = tf.data.Dataset.from_generator(
            lambda: self, output_signature=self.make_signature()
        )

        ds = ds.shuffle(
            buffer_size=self.buffer_size, reshuffle_each_iteration=True
        ).batch(batch_size=self.batch_size)
        if self.n_jit_steps > 1:
            ds = ds.batch(batch_size=self.n_jit_steps)
        ds = prefetch_to_single_device(
            ds.as_numpy_iterator(), 2, sharding, n_step_jit=self.n_jit_steps > 1
        )
        return ds

    def batch(self, sharding=None) -> Iterator[jax.Array]:
        ds = tf.data.Dataset.from_generator(
            lambda: self, output_signature=self.make_signature()
        )
        ds = ds.batch(batch_size=self.batch_size)
        ds = prefetch_to_single_device(
            ds.as_numpy_iterator(), 2, sharding, n_step_jit=self.n_jit_steps > 1
        )
        return ds


def next_power_of_two(x):
    return 1 << (int(x) - 1).bit_length()


def round_up_to_multiple(value, multiple):
    """
    Rounds up the given integer `value` to the next multiple of `multiple`.

    Parameters:
    - value (int): The integer to round up.
    - multiple (int): The multiple to round up to.

    Returns:
    - int: The rounded-up value.
    """
    return int(np.ceil(value / multiple) * multiple)


class BatchProcessor:
    def __init__(
        self,
        cutoff,
        atom_padding: int,
        nl_padding: int,
        forces=True,
        stress=False,
        additional_properties=[],
    ) -> None:
        self.cutoff = cutoff
        self.atom_padding = atom_padding
        self.nl_padding = nl_padding

        self.forces = forces
        self.stress = stress
        self.additional_properties = additional_properties

    def __call__(self, samples: list[dict]):
        n_samples = len(samples)
        max_atoms = np.max([inp[0]["n_atoms"] for inp in samples])
        max_atoms = round_up_to_multiple(max_atoms, self.atom_padding)

        inputs = {
            "numbers": np.zeros((n_samples, max_atoms), dtype=np.int16),
            "n_atoms": np.zeros(n_samples, dtype=np.int16),
            "positions": np.zeros((n_samples, max_atoms, 3), dtype=np.float64),
            "box": np.zeros((n_samples, 3, 3), dtype=np.float32),
        }

        labels = {
            "energy": np.zeros(n_samples, dtype=np.float64),
        }
        for prop in self.additional_properties:
            name, shape = prop
            if shape[0] == "natoms":
                shape = [max_atoms] + shape[1:]
            shape = [n_samples] + shape
            labels[name] = np.zeros(shape, dtype=np.float64)
        if self.forces:
            labels["forces"] = np.zeros((n_samples, max_atoms, 3), dtype=np.float64)
        if self.stress:
            labels["stress"] = np.zeros((n_samples, 3, 3), dtype=np.float64)

        idxs = []
        offsets = []
        for i, (inp, lab) in enumerate(samples):
            inputs["numbers"][i, : inp["n_atoms"]] = inp["numbers"]
            inputs["n_atoms"][i] = inp["n_atoms"]
            inputs["positions"][i, : inp["n_atoms"]] = inp["positions"]
            inputs["box"][i] = inp["box"]

            idx, offset = compute_nl(inp["positions"], inp["box"], self.cutoff)
            idxs.append(idx)
            offsets.append(offset)

            labels["energy"][i] = lab["energy"]
            if self.forces:
                labels["forces"][i, : inp["n_atoms"]] = lab["forces"]
            if self.stress:
                labels["stress"][i] = lab["stress"]

            for prop in self.additional_properties:
                name, shape = prop
                if shape[0] == "natoms":
                    labels[name][i, : inp["n_atoms"]] = lab[name]
                else:
                    labels[name][i] = lab[name]

        max_nbrs = np.max([idx.shape[1] for idx in idxs])
        max_nbrs = round_up_to_multiple(max_nbrs, self.nl_padding)

        inputs["idx"] = np.zeros((n_samples, 2, max_nbrs), dtype=np.int16)
        inputs["offsets"] = np.zeros((n_samples, max_nbrs, 3), dtype=np.float64)

        for i, (idx, offset) in enumerate(zip(idxs, offsets)):
            inputs["idx"][i, :, : idx.shape[1]] = idx
            inputs["offsets"][i, : offset.shape[0], :] = offset

        return inputs, labels


class PerBatchPaddedDataset(InMemoryDataset):
    """Dataset with padding that leverages multiprocessing and optimized buffering.

    Per-atom and per-neighbor arrays are padded to the next multiple of a user specified integer.
    This limits the compute wasted due to padding at the (negligible) cost of some recompilations.
    Since the padding occurs on a per-batch basis, it is the most performant option for datasets with significantly differently sized systems (e.g. MaterialsProject, SPICE).

    Further, the neighborlist is computed on-the-fly in parallel on a side thread.
    Does not use tf.data.

    Attributes
    ----------
    num_workers : int
        Number of processes to use for preprocessing batches.
    atom_padding : int
        Pad extensive arrays (positions, etc.) to next multiple of this integer.
    nl_padding : int
        Pad neighborlist arrays to next multiple of this integer.
    """

    def __init__(
        self,
        atoms_list,
        cutoff,
        bs,
        n_epochs,
        n_jit_steps=1,
        num_workers: Optional[int] = None,
        atom_padding: int = 10,
        nl_padding: int = 2000,
        pos_unit: str = "Ang",
        energy_unit: str = "eV",
        additional_properties=[],
        pre_shuffle=False,
    ) -> None:
        self.cutoff = cutoff
        self.n_jit_steps = n_jit_steps
        self.n_epochs = n_epochs
        self.n_data = len(atoms_list)
        self.batch_size = self.validate_batch_size(bs)
        self.pos_unit = pos_unit
        self.additional_properties = additional_properties

        if num_workers:
            self.num_workers = num_workers
        else:
            self.num_workers = multiprocessing.cpu_count() - 1

        self.sample_atoms = atoms_list[0]

        # Transform atoms into inputs and labels
        self.inputs = atoms_to_inputs(atoms_list, pos_unit)
        self.labels = atoms_to_labels(
            atoms_list, pos_unit, energy_unit, additional_properties
        )
        label_keys = self.labels.keys()

        self.data = list(
            zip(
                transpose_dict_of_lists(self.inputs), transpose_dict_of_lists(self.labels)
            )
        )

        forces = "forces" in label_keys
        stress = "stress" in label_keys
        self.prepare_batch = BatchProcessor(
            cutoff, atom_padding, nl_padding, forces, stress, additional_properties
        )

        self.count = 0

        self.max_count = self.n_epochs * self.steps_per_epoch()

        self.buffer_size = min(600, self.steps_per_epoch())
        self.buffer = Queue(maxsize=self.buffer_size)

        self.process_pool = ProcessPoolExecutor(self.num_workers)
        self.thread_pool = ThreadPoolExecutor(1)  # Single thread for buffering batches
        self.epoch_finished = False
        self.enqueue_future = None
        self.needs_data = Event()

    def enqueue_batches(self):
        """Function to enqueue batches on a side thread."""
        while self.count < self.steps_per_epoch() * self.n_epochs:
            self.needs_data.wait()
            if self.epoch_finished:
                break
            num_batches = min(
                self.buffer_size - self.buffer.qsize(),
                self.steps_per_epoch() - self.count,
            )
            if num_batches > 0:
                self.enqueue(num_batches)
            self.needs_data.clear()  # Reset event

    def enqueue(self, num_batches):
        start = self.count * self.batch_size

        # Split data into chunks and submit tasks to the process pool
        dataset_chunks = [
            self.data[start + self.batch_size * i : start + self.batch_size * (i + 1)]
            for i in range(num_batches)
        ]

        # Using submit and as_completed for faster batch retrieval
        futures = [
            self.process_pool.submit(self.prepare_batch, chunk)
            for chunk in dataset_chunks
        ]
        for future in as_completed(futures):
            batch = future.result()
            self.buffer.put(batch)

        self.count += num_batches

    def __iter__(self):
        for n in range(self.n_epochs):
            self.count = 0
            self.buffer.queue.clear()  # Reset buffer
            self.epoch_finished = False

            if self.should_shuffle:
                shuffle(self.data)

            # Start pre-filling the buffer
            self.enqueue_future = self.thread_pool.submit(self.enqueue_batches)

            for i in range(self.steps_per_epoch()):
                if self.buffer.qsize() < (self.buffer_size * 0.75):
                    self.needs_data.set()  # Trigger buffer refill
                while self.buffer.empty():
                    time.sleep(0.001)
                yield self.buffer.get()

            self.epoch_finished = True
            self.needs_data.set()
            self.enqueue_future.result()

    def shuffle_and_batch(self, sharding):
        self.should_shuffle = True
        ds = prefetch_to_single_device(
            iter(self), 2, sharding, n_step_jit=self.n_jit_steps > 1
        )
        return ds

    def batch(self, sharding) -> Iterator[jax.Array]:
        self.should_shuffle = False
        ds = prefetch_to_single_device(
            iter(self), 2, sharding, n_step_jit=self.n_jit_steps > 1
        )
        return ds

    def make_signature(self) -> None:
        pass

    def cleanup(self):
        self.epoch_finished = True
        self.needs_data.set()
        self.enqueue_future.result()
        self.needs_data.clear()
        self.thread_pool.shutdown(wait=True, cancel_futures=True)
        self.process_pool.shutdown(wait=True, cancel_futures=True)
        self.buffer.queue.clear()


dataset_dict = {
    "cached": CachedInMemoryDataset,
    "otf": OTFInMemoryDataset,
    "pbp": PerBatchPaddedDataset,
}



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code to simulate systems in various statistical ensembles.

This file contains a number of different methods that can be used to
simulate systems in a variety of ensembles.

In general, simulation code follows the same overall structure as optimizers
in JAX. Simulations are tuples of two functions:

  init_fn:
    Function that initializes the  state of a system. Should take
    positions as an ndarray of shape `[n, output_dimension]`. Returns a state
    which will be a namedtuple.
  apply_fn:
    Function that takes a state and produces a new state after one
    step of optimization.

One question that we need to think about is whether the simulations should
also return a function that computes the invariant for that ensemble. This
can be used for testing purposes, but is not often used otherwise.
"""

from typing import Callable, Dict, Optional, Tuple, TypeVar

import jax.numpy as jnp
from jax import grad, jit, lax, random, tree_util

from apax.utils.jax_md_reduced import (
    dataclasses,
    partition,
    quantity,
    smap,
    space,
    util,
)

static_cast = util.static_cast


# Types


Array = util.Array
f32 = util.f32
f64 = util.f64

Box = space.Box

ShiftFn = space.ShiftFn

T = TypeVar("T")
InitFn = Callable[..., T]
ApplyFn = Callable[[T], T]
Simulator = Tuple[InitFn, ApplyFn]


"""Dispatch By State Code.

JAX MD allows for simulations to be extensible using a dispatch strategy where
functions are dispatched to specific cases based on the type of state provided.
In particular, we make decisions about which function to call based on the type
of the position argument. For those familiar with C / C++, our dispatch code is
essentially function overloading based on the type of the positions.

If you are interested in setting up a simulation using a different type of
system you can do so in a relatively light weight manner by introducing a new
type for storing the state that is compatible with the JAX PyTree system
(we usually choose a dataclass) and then overriding the functions below.

These extensions allow a range of simulations to be run by just changing the
type of the position argument. There are essentially two types of functions to
be overloaded. Functions that compute physical quantities, such as the kinetic
energy, and functions that evolve a state according to the Suzuki-Trotter
decomposition. Specifically, one might want to override the position step,
momentum step for deterministic and stochastic simulations or the
`stochastic_step` for stochastic simulations (e.g Langevin).
"""


class dispatch_by_state:
    """Wrap a function and dispatch based on the type of positions."""

    def __init__(self, fn):
        self._fn = fn
        self._registry = {}

    def __call__(self, state, *args, **kwargs):
        if type(state.position) in self._registry:
            return self._registry[type(state.position)](state, *args, **kwargs)
        return self._fn(state, *args, **kwargs)

    def register(self, oftype):
        def register_fn(fn):
            self._registry[oftype] = fn

        return register_fn


@dispatch_by_state
def canonicalize_mass(state: T) -> T:
    """Reshape mass vector for broadcasting with positions."""

    def canonicalize_fn(mass):
        if isinstance(mass, float):
            return mass
        if mass.ndim == 2 and mass.shape[1] == 1:
            return mass
        elif mass.ndim == 1:
            return jnp.reshape(mass, (mass.shape[0], 1))
        elif mass.ndim == 0:
            return mass
        msg = (
            "Expected mass to be either a floating point number or a one-dimensional"
            "ndarray. Found {}.".format(mass)
        )
        raise ValueError(msg)

    return state.set(mass=tree_util.tree_map(canonicalize_fn, state.mass))


@dispatch_by_state
def initialize_momenta(state: T, key: Array, kT: float) -> T:
    """Initialize momenta with the Maxwell-Boltzmann distribution."""
    R, mass = state.position, state.mass

    R, treedef = tree_util.tree_flatten(R)
    mass, _ = tree_util.tree_flatten(mass)
    keys = random.split(key, len(R))

    def initialize_fn(k, r, m):
        p = jnp.sqrt(m * kT) * random.normal(k, r.shape, dtype=r.dtype)
        # If simulating more than one particle, center the momentum.
        if r.shape[0] > 1:
            p = p - jnp.mean(p, axis=0, keepdims=True)
        return p

    P = [initialize_fn(k, r, m) for k, r, m in zip(keys, R, mass)]

    return state.set(momentum=tree_util.tree_unflatten(treedef, P))


@dispatch_by_state
def momentum_step(state: T, dt: float) -> T:
    """Apply a single step of the time evolution operator for momenta."""
    assert hasattr(state, "momentum")
    new_momentum = tree_util.tree_map(
        lambda p, f: p + dt * f, state.momentum, state.force
    )
    return state.set(momentum=new_momentum)


@dispatch_by_state
def position_step(state: T, shift_fn: Callable, dt: float, **kwargs) -> T:
    """Apply a single step of the time evolution operator for positions."""
    if isinstance(shift_fn, Callable):
        shift_fn = tree_util.tree_map(lambda r: shift_fn, state.position)
    new_position = tree_util.tree_map(
        lambda s_fn, r, p, m: s_fn(r, dt * p / m, **kwargs),
        shift_fn,
        state.position,
        state.momentum,
        state.mass,
    )
    return state.set(position=new_position)


@dispatch_by_state
def kinetic_energy(state: T) -> Array:
    """Compute the kinetic energy of a state."""
    return quantity.kinetic_energy(momentum=state.momentum, mass=state.mass)


@dispatch_by_state
def temperature(state: T) -> Array:
    """Compute the temperature of a state."""
    return quantity.temperature(momentum=state.momentum, mass=state.mass)


"""Deterministic Simulations

JAX MD includes integrators for deterministic simulations of the NVE, NVT, and
NPT ensembles. For a qualitative description of statistical physics ensembles
see the wikipedia article here:
en.wikipedia.org/wiki/Statistical_ensemble_(mathematical_physics)

Integrators are based direct translation method outlined in the paper,

"A Liouville-operator derived measure-preserving integrator for molecular
dynamics simulations in the isothermal–isobaric ensemble"

M. E. Tuckerman, J. Alejandre, R. López-Rendón, A. L Jochim, and G. J. Martyna
J. Phys. A: Math. Gen. 39 5629 (2006)

As such, we define several primitives that are generically useful in describing
simulations of this type. Namely, the velocity-Verlet integration step that is
used in the NVE and NVT simulations. We also define a general Nose-Hoover chain
primitive that is used to couple components of the system to a chain that
regulates the temperature. These primitives can be combined to construct more
interesting simulations that involve e.g. temperature gradients.
"""


def velocity_verlet(
    force_fn: Callable[..., Array], shift_fn: ShiftFn, dt: float, state: T, **kwargs
) -> T:
    """Apply a single step of velocity Verlet integration to a state."""
    dt = f32(dt)
    dt_2 = f32(dt / 2)

    state = momentum_step(state, dt_2)
    state = position_step(state, shift_fn, dt, **kwargs)
    state = state.set(force=force_fn(state.position, **kwargs))
    state = momentum_step(state, dt_2)

    return state


# Constant Energy Simulations


@dataclasses.dataclass
class NVEState:
    """A struct containing the state of an NVE simulation.

    This tuple stores the state of a simulation that samples from the
    microcanonical ensemble in which the (N)umber of particles, the (V)olume, and
    the (E)nergy of the system are held fixed.

    Attributes:
      position: An ndarray of shape `[n, spatial_dimension]` storing the position
        of particles.
      momentum: An ndarray of shape `[n, spatial_dimension]` storing the momentum
        of particles.
      force: An ndarray of shape `[n, spatial_dimension]` storing the force
        acting on particles from the previous step.
      mass: A float or an ndarray of shape `[n]` containing the masses of the
        particles.
    """

    position: Array
    momentum: Array
    force: Array
    mass: Array

    @property
    def velocity(self) -> Array:
        return self.momentum / self.mass


# pylint: disable=invalid-name
def nve(energy_or_force_fn, shift_fn, kT, dt=1e-3, **sim_kwargs):
    """Simulates a system in the NVE ensemble.

    Samples from the microcanonical ensemble in which the number of particles
    (N), the system volume (V), and the energy (E) are held constant. We use a
    standard velocity Verlet integration scheme.

    Args:
      energy_or_force: A function that produces either an energy or a force from
        a set of particle positions specified as an ndarray of shape
        `[n, spatial_dimension]`.
      shift_fn: A function that displaces positions, `R`, by an amount `dR`.
        Both `R` and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
      dt: Floating point number specifying the timescale (step size) of the
        simulation.
    Returns:
      See above.
    """
    force_fn = quantity.canonicalize_force(energy_or_force_fn)

    @jit
    def init_fn(key, R, mass=f32(1.0), **kwargs):
        force = force_fn(R, **kwargs)
        state = NVEState(R, None, force, mass)
        state = canonicalize_mass(state)
        return initialize_momenta(state, key, kT)

    @jit
    def step_fn(state, **kwargs):
        _dt = kwargs.pop("dt", dt)
        _ = kwargs.pop("kT")
        return velocity_verlet(force_fn, shift_fn, _dt, state, **kwargs)

    return init_fn, step_fn


# Constant Temperature Simulations


# Suzuki-Yoshida weights for integrators of different order.
# These are copied from OpenMM at
# https://github.com/openmm/openmm/blob/master/openmmapi/src/NoseHooverChain.cpp


SUZUKI_YOSHIDA_WEIGHTS = {
    1: [1],
    3: [0.828981543588751, -0.657963087177502, 0.828981543588751],
    5: [
        0.2967324292201065,
        0.2967324292201065,
        -0.186929716880426,
        0.2967324292201065,
        0.2967324292201065,
    ],
    7: [
        0.784513610477560,
        0.235573213359357,
        -1.17767998417887,
        1.31518632068391,
        -1.17767998417887,
        0.235573213359357,
        0.784513610477560,
    ],
}


@dataclasses.dataclass
class NoseHooverChain:
    """State information for a Nose-Hoover chain.

    Attributes:
      position: An ndarray of shape `[chain_length]` that stores the position of
        the chain.
      momentum: An ndarray of shape `[chain_length]` that stores the momentum of
        the chain.
      mass: An ndarray of shape `[chain_length]` that stores the mass of the
        chain.
      tau: The desired period of oscillation for the chain. Longer periods result
        is better stability but worse temperature control.
      kinetic_energy: A float that stores the current kinetic energy of the
        system that the chain is coupled to.
      degrees_of_freedom: An integer specifying the number of degrees of freedom
        that the chain is coupled to.
    """

    position: Array
    momentum: Array
    mass: Array
    tau: Array
    kinetic_energy: Array
    degrees_of_freedom: int = dataclasses.static_field()


@dataclasses.dataclass
class NoseHooverChainFns:
    initialize: Callable
    half_step: Callable
    update_mass: Callable


def nose_hoover_chain(
    dt: float, chain_length: int, chain_steps: int, sy_steps: int, tau: float
) -> NoseHooverChainFns:
    """Helper function to simulate a Nose-Hoover Chain coupled to a system.

    This function is used in simulations that sample from thermal ensembles by
    coupling the system to one, or more, Nose-Hoover chains. We use the direct
    translation method outlined in Martyna et al. [#martyna92]_ and the
    Nose-Hoover chains are updated using two half steps: one at the beginning of
    a simulation step and one at the end. The masses of the Nose-Hoover chains
    are updated automatically to enforce a specific period of oscillation, `tau`.
    Larger values of `tau` will yield systems that reach the target temperature
    more slowly but are also more stable.

    As described in Martyna et al. [#martyna92]_, the Nose-Hoover chain often
    evolves on a faster timescale than the rest of the simulation. Therefore, it
    sometimes necessary
    to integrate the chain over several substeps for each step of MD. To do this
    we follow the Suzuki-Yoshida scheme. Specifically, we subdivide our chain
    simulation into :math:`n_c` substeps. These substeps are further subdivided
    into :math:`n_sy` steps. Each :math:`n_sy` step has length
    :math:`\delta_i = \Delta t w_i / n_c` where :math:`w_i` are constants such
    that :math:`\sum_i w_i = 1`. See the table of Suzuki-Yoshida weights above
    for specific values. The number of substeps and the number of Suzuki-Yoshida
    steps are set using the `chain_steps` and `sy_steps` arguments.

    Consequently, the Nose-Hoover chains are described by three functions: an
    `init_fn` that initializes the state of the chain, a `half_step_fn` that
    updates the chain for one half-step, and an `update_chain_mass_fn` that
    updates the masses of the chain to enforce the correct period of oscillation.

    Note that a system can have many Nose-Hoover chains coupled to it to produce,
    for example, a temperature gradient. We also note that the NPT ensemble
    naturally features two chains: one that couples to the thermal degrees of
    freedom and one that couples to the barostat.

    Attributes:
      dt: Floating point number specifying the timescale (step size) of the
        simulation.
      chain_length: An integer specifying the number of particles in
        the Nose-Hoover chain.
      chain_steps: An integer specifying the number :math:`n_c` of outer substeps.
      sy_steps: An integer specifying the number of Suzuki-Yoshida steps. This
        must be either `1`, `3`, `5`, or `7`.
      tau: A floating point timescale over which temperature equilibration occurs.
        Measured in units of `dt`. The performance of the Nose-Hoover chain
        thermostat can be quite sensitive to this choice.
    Returns:
      A triple of functions that initialize the chain, do a half step of
      simulation, and update the chain masses respectively.
    """

    def init_fn(degrees_of_freedom, KE, kT):
        xi = jnp.zeros(chain_length, KE.dtype)
        p_xi = jnp.zeros(chain_length, KE.dtype)

        Q = kT * tau ** f32(2) * jnp.ones(chain_length, dtype=f32)
        Q = Q.at[0].multiply(degrees_of_freedom)
        return NoseHooverChain(xi, p_xi, Q, tau, KE, degrees_of_freedom)

    def substep_fn(delta, P, state, kT):
        """Apply a single update to the chain parameters and rescales velocity."""
        xi, p_xi, Q, _tau, KE, DOF = dataclasses.astuple(state)

        delta_2 = delta / f32(2.0)
        delta_4 = delta_2 / f32(2.0)
        delta_8 = delta_4 / f32(2.0)

        M = chain_length - 1

        G = p_xi[M - 1] ** f32(2) / Q[M - 1] - kT
        p_xi = p_xi.at[M].add(delta_4 * G)

        def backward_loop_fn(p_xi_new, m):
            G = p_xi[m - 1] ** 2 / Q[m - 1] - kT
            scale = jnp.exp(-delta_8 * p_xi_new / Q[m + 1])
            p_xi_new = scale * (scale * p_xi[m] + delta_4 * G)
            return p_xi_new, p_xi_new

        idx = jnp.arange(M - 1, 0, -1)
        _, p_xi_update = lax.scan(backward_loop_fn, p_xi[M], idx, unroll=2)
        p_xi = p_xi.at[idx].set(p_xi_update)

        G = f32(2.0) * KE - DOF * kT
        scale = jnp.exp(-delta_8 * p_xi[1] / Q[1])
        p_xi = p_xi.at[0].set(scale * (scale * p_xi[0] + delta_4 * G))

        scale = jnp.exp(-delta_2 * p_xi[0] / Q[0])
        KE = KE * scale ** f32(2)
        P = tree_util.tree_map(lambda p: p * scale, P)

        xi = xi + delta_2 * p_xi / Q

        G = f32(2) * KE - DOF * kT

        def forward_loop_fn(G, m):
            scale = jnp.exp(-delta_8 * p_xi[m + 1] / Q[m + 1])
            p_xi_update = scale * (scale * p_xi[m] + delta_4 * G)
            G = p_xi_update**2 / Q[m] - kT
            return G, p_xi_update

        idx = jnp.arange(M)
        G, p_xi_update = lax.scan(forward_loop_fn, G, idx, unroll=2)
        p_xi = p_xi.at[idx].set(p_xi_update)
        p_xi = p_xi.at[M].add(delta_4 * G)

        return P, NoseHooverChain(xi, p_xi, Q, _tau, KE, DOF), kT

    def half_step_chain_fn(P, state, kT):
        if chain_steps == 1 and sy_steps == 1:
            P, state, _ = substep_fn(dt, P, state, kT)
            return P, state

        delta = dt / chain_steps
        ws = jnp.array(SUZUKI_YOSHIDA_WEIGHTS[sy_steps])

        def body_fn(cs, i):
            d = f32(delta * ws[i % sy_steps])
            return substep_fn(d, *cs), 0

        P, state, _ = lax.scan(
            body_fn, (P, state, kT), jnp.arange(chain_steps * sy_steps)
        )[0]
        return P, state

    def update_chain_mass_fn(state, kT):
        xi, p_xi, Q, _tau, KE, DOF = dataclasses.astuple(state)

        Q = kT * _tau ** f32(2) * jnp.ones(chain_length, dtype=f32)
        Q = Q.at[0].multiply(DOF)

        return NoseHooverChain(xi, p_xi, Q, _tau, KE, DOF)

    return NoseHooverChainFns(init_fn, half_step_chain_fn, update_chain_mass_fn)


def default_nhc_kwargs(tau: float, overrides: Dict) -> Dict:
    default_kwargs = {"chain_length": 3, "chain_steps": 2, "sy_steps": 3, "tau": tau}

    if overrides is None:
        return default_kwargs

    return {key: overrides.get(key, default_kwargs[key]) for key in default_kwargs}


@dataclasses.dataclass
class NVTNoseHooverState:
    """State information for an NVT system with a Nose-Hoover chain thermostat.

    Attributes:
      position: The current position of particles. An ndarray of floats
        with shape `[n, spatial_dimension]`.
      momentum: The momentum of particles. An ndarray of floats
        with shape `[n, spatial_dimension]`.
      force: The current force on the particles. An ndarray of floats with shape
        `[n, spatial_dimension]`.
      mass: The mass of the particles. Can either be a float or an ndarray
        of floats with shape `[n]`.
      chain: The variables describing the Nose-Hoover chain.
    """

    position: Array
    momentum: Array
    force: Array
    mass: Array
    chain: NoseHooverChain

    @property
    def velocity(self):
        return self.momentum / self.mass


def nvt_nose_hoover(
    energy_or_force_fn: Callable[..., Array],
    shift_fn: ShiftFn,
    dt: float,
    kT: float,
    chain_length: int = 5,
    chain_steps: int = 2,
    sy_steps: int = 3,
    tau: Optional[float] = None,
    **sim_kwargs,
) -> Simulator:
    """Simulation in the NVT ensemble using a Nose Hoover Chain thermostat.

    Samples from the canonical ensemble in which the number of particles (N),
    the system volume (V), and the temperature (T) are held constant. We use a
    Nose Hoover Chain (NHC) thermostat described in [#martyna92]_ [#martyna98]_
    [#tuckerman]_. We follow the direct translation method outlined in
    Tuckerman et al. [#tuckerman]_ and the interested reader might want to look
    at that paper as a reference.

    Args:
      energy_or_force: A function that produces either an energy or a force from
        a set of particle positions specified as an ndarray of shape
        `[n, spatial_dimension]`.
      shift_fn: A function that displaces positions, `R`, by an amount `dR`.
        Both `R` and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
      dt: Floating point number specifying the timescale (step size) of the
        simulation.
      kT: Floating point number specifying the temperature in units of Boltzmann
        constant. To update the temperature dynamically during a simulation one
        should pass `kT` as a keyword argument to the step function.
      chain_length: An integer specifying the number of particles in
        the Nose-Hoover chain.
      chain_steps: An integer specifying the number, :math:`n_c`, of outer
        substeps.
      sy_steps: An integer specifying the number of Suzuki-Yoshida steps. This
        must be either `1`, `3`, `5`, or `7`.
      tau: A floating point timescale over which temperature equilibration
        occurs. Measured in units of `dt`. The performance of the Nose-Hoover
        chain thermostat can be quite sensitive to this choice.
    Returns:
      See above.

    .. rubric:: References
    .. [#martyna92] Martyna, Glenn J., Michael L. Klein, and Mark Tuckerman.
      "Nose-Hoover chains: The canonical ensemble via continuous dynamics."
      The Journal of chemical physics 97, no. 4 (1992): 2635-2643.
    .. [#martyna98] Martyna, Glenn, Mark Tuckerman, Douglas J. Tobias, and Michael L. Klein.
      "Explicit reversible integrators for extended systems dynamics."
      Molecular Physics 87. (1998) 1117-1157.
    .. [#tuckerman] Tuckerman, Mark E., Jose Alejandre, Roberto Lopez-Rendon,
      Andrea L. Jochim, and Glenn J. Martyna.
      "A Liouville-operator derived measure-preserving integrator for molecular
      dynamics simulations in the isothermal-isobaric ensemble."
      Journal of Physics A: Mathematical and General 39, no. 19 (2006): 5629.
    """
    force_fn = quantity.canonicalize_force(energy_or_force_fn)
    dt = f32(dt)
    dt_2 = f32(dt / 2)
    if tau is None:
        tau = dt * 100
    tau = f32(tau)

    thermostat = nose_hoover_chain(dt, chain_length, chain_steps, sy_steps, tau)

    @jit
    def init_fn(key, R, mass=f32(1.0), **kwargs):
        _kT = kT if "kT" not in kwargs else kwargs["kT"]

        dof = quantity.count_dof(R)

        state = NVTNoseHooverState(R, None, force_fn(R, **kwargs), mass, None)
        state = canonicalize_mass(state)
        state = initialize_momenta(state, key, _kT)
        KE = kinetic_energy(state)
        return state.set(chain=thermostat.initialize(dof, KE, _kT))

    @jit
    def apply_fn(state, **kwargs):
        if "kT" not in kwargs:
            _kT = kT
        else:
            _kT = kwargs.pop("kT")

        chain = state.chain

        chain = thermostat.update_mass(chain, _kT)

        p, chain = thermostat.half_step(state.momentum, chain, _kT)
        state = state.set(momentum=p)

        state = velocity_verlet(force_fn, shift_fn, dt, state, **kwargs)

        chain = chain.set(kinetic_energy=kinetic_energy(state))

        p, chain = thermostat.half_step(state.momentum, chain, _kT)
        state = state.set(momentum=p, chain=chain)

        return state

    return init_fn, apply_fn


def nvt_nose_hoover_invariant(
    energy_fn: Callable[..., Array], state: NVTNoseHooverState, kT: float, **kwargs
) -> float:
    """The conserved quantity for the NVT ensemble with a Nose-Hoover thermostat.

    This function is normally used for debugging the Nose-Hoover thermostat.

    Arguments:
      energy_fn: The energy function of the Nose-Hoover system.
      state: The current state of the system.
      kT: The current goal temperature of the system.

    Returns:
      The Hamiltonian of the extended NVT dynamics.
    """
    PE = energy_fn(state.position, **kwargs)
    KE = kinetic_energy(state)

    DOF = quantity.count_dof(state.position)
    E = PE + KE

    c = state.chain

    E += c.momentum[0] ** 2 / (2 * c.mass[0]) + DOF * kT * c.position[0]
    for r, p, m in zip(c.position[1:], c.momentum[1:], c.mass[1:]):
        E += p**2 / (2 * m) + kT * r
    return E


@dataclasses.dataclass
class NPTNoseHooverState:
    """State information for an NPT system with Nose-Hoover chain thermostats.

    Attributes:
      position: The current position of particles. An ndarray of floats
        with shape `[n, spatial_dimension]`.
      momentum: The velocity of particles. An ndarray of floats
        with shape `[n, spatial_dimension]`.
      force: The current force on the particles. An ndarray of floats with shape
        `[n, spatial_dimension]`.
      mass: The mass of the particles. Can either be a float or an ndarray
        of floats with shape `[n]`.
      reference_box: A box used to measure relative changes to the simulation
        environment.
      box_position: A positional degree of freedom used to describe the current
        box. box_position is parameterized as `box_position = (1/d)log(V/V_0)`
        where `V` is the current volume, `V_0` is the reference volume, and `d`
        is the spatial dimension.
      box_velocity: A velocity degree of freedom for the box.
      box_mass: The mass assigned to the box.
      barostat: The variables describing the Nose-Hoover chain coupled to the
        barostat.
      thermostsat: The variables describing the Nose-Hoover chain coupled to the
        thermostat.
    """

    position: Array
    momentum: Array
    force: Array
    mass: Array

    reference_box: Box

    box_position: Array
    box_momentum: Array
    box_mass: Array

    barostat: NoseHooverChain
    thermostat: NoseHooverChain

    @property
    def velocity(self) -> Array:
        return self.momentum / self.mass

    @property
    def box(self) -> Array:
        """Get the current box from an NPT simulation."""
        dim = self.position.shape[1]
        ref = self.reference_box
        V_0 = quantity.volume(dim, ref)
        V = V_0 * jnp.exp(dim * self.box_position)
        return (V / V_0) ** (1 / dim) * ref


def _npt_box_info(state: NPTNoseHooverState) -> Tuple[float, Callable[[float], float]]:
    """Gets the current volume and a function to compute the box from volume."""
    dim = state.position.shape[1]
    ref = state.reference_box
    V_0 = quantity.volume(dim, ref)
    V = V_0 * jnp.exp(dim * state.box_position)
    return V, lambda V: (V / V_0) ** (1 / dim) * ref


def npt_box(state: NPTNoseHooverState) -> Box:
    """Get the current box from an NPT simulation."""
    dim = state.position.shape[1]
    ref = state.reference_box
    V_0 = quantity.volume(dim, ref)
    V = V_0 * jnp.exp(dim * state.box_position)
    return (V / V_0) ** (1 / dim) * ref


def npt_nose_hoover(
    energy_fn: Callable[..., Array],
    shift_fn: ShiftFn,
    dt: float,
    pressure: float,
    kT: float,
    barostat_kwargs: Optional[Dict] = None,
    thermostat_kwargs: Optional[Dict] = None,
) -> Simulator:
    """Simulation in the NPT ensemble using a pair of Nose Hoover Chains.

    Samples from the canonical ensemble in which the number of particles (N),
    the system pressure (P), and the temperature (T) are held constant.
    We use a pair of Nose Hoover Chains (NHC) described in
    [#martyna92]_ [#martyna98]_ [#tuckerman]_ coupled to the
    barostat and the thermostat respectively. We follow the direct translation
    method outlined in Tuckerman et al. [#tuckerman]_ and the interested reader
    might want to look at that paper as a reference.

    Args:
      energy_fn: A function that produces either an energy from a set of particle
        positions specified as an ndarray of shape `[n, spatial_dimension]`.
      shift_fn: A function that displaces positions, `R`, by an amount `dR`. Both
        `R` and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
      dt: Floating point number specifying the timescale (step size) of the
        simulation.
      pressure: Floating point number specifying the target pressure. To update
        the pressure dynamically during a simulation one should pass `pressure`
        as a keyword argument to the step function.
      kT: Floating point number specifying the temperature in units of Boltzmann
        constant. To update the temperature dynamically during a simulation one
        should pass `kT` as a keyword argument to the step function.
      barostat_kwargs: A dictionary of keyword arguments passed to the barostat
        NHC. Any parameters not set are drawn from a relatively robust default
        set.
      thermostat_kwargs: A dictionary of keyword arguments passed to the
        thermostat NHC. Any parameters not set are drawn from a relatively robust
        default set.

    Returns:
      See above.

    """

    t = f32(dt)
    dt_2 = f32(dt / 2)

    force_fn = quantity.force(energy_fn)

    barostat_kwargs = default_nhc_kwargs(1000 * dt, barostat_kwargs)
    barostat = nose_hoover_chain(dt, **barostat_kwargs)

    thermostat_kwargs = default_nhc_kwargs(100 * dt, thermostat_kwargs)
    thermostat = nose_hoover_chain(dt, **thermostat_kwargs)

    def init_fn(key, R, box, mass=f32(1.0), **kwargs):
        N, dim = R.shape

        if "kT" not in kwargs:
            _kT = kT
        else:
            _kT = kwargs.pop("kT")

        # The box position is defined via pos = (1 / d) log V / V_0.
        zero = jnp.zeros((), dtype=R.dtype)
        one = jnp.ones((), dtype=R.dtype)
        box_position = zero
        box_momentum = zero
        box_mass = dim * (N + 1) * kT * barostat_kwargs["tau"] ** 2 * one
        KE_box = quantity.kinetic_energy(momentum=box_momentum, mass=box_mass)

        if jnp.isscalar(box) or box.ndim == 0:
            # TODO(schsam): This is necessary because of JAX issue #5849.
            box = jnp.eye(R.shape[-1]) * box

        state = NPTNoseHooverState(
            R,
            None,
            force_fn(R, box=box, **kwargs),
            mass,
            box,
            box_position,
            box_momentum,
            box_mass,
            barostat.initialize(1, KE_box, _kT),
            None,
        )  # pytype: disable=wrong-arg-count
        state = canonicalize_mass(state)
        state = initialize_momenta(state, key, _kT)
        KE = kinetic_energy(state)
        return state.set(thermostat=thermostat.initialize(quantity.count_dof(R), KE, _kT))

    def update_box_mass(state, kT):
        N, dim = state.position.shape
        dtype = state.position.dtype
        box_mass = jnp.array(dim * (N + 1) * kT * state.barostat.tau**2, dtype)
        return state.set(box_mass=box_mass)

    def box_force(
        alpha, vol, box_fn, position, momentum, mass, force, pressure, **kwargs
    ):
        N, dim = position.shape

        def U(eps):
            return energy_fn(position, box=box_fn(vol), perturbation=(1 + eps), **kwargs)

        dUdV = grad(U)
        KE2 = util.high_precision_sum(momentum**2 / mass)

        return alpha * KE2 - dUdV(0.0) - pressure * vol * dim

    def sinhx_x(x):
        """Taylor series for sinh(x) / x as x -> 0."""
        return (
            1 + x**2 / 6 + x**4 / 120 + x**6 / 5040 + x**8 / 362_880 + x**10 / 39_916_800
        )

    def exp_iL1(box, R, V, V_b, **kwargs):
        x = V_b * dt
        x_2 = x / 2
        sinhV = sinhx_x(x_2)  # jnp.sinh(x_2) / x_2
        return shift_fn(
            R, R * (jnp.exp(x) - 1) + dt * V * jnp.exp(x_2) * sinhV, box=box, **kwargs
        )  # pytype: disable=wrong-keyword-args

    def exp_iL2(alpha, P, F, V_b):
        x = alpha * V_b * dt_2
        x_2 = x / 2
        sinhP = sinhx_x(x_2)  # jnp.sinh(x_2) / x_2
        return P * jnp.exp(-x) + dt_2 * F * sinhP * jnp.exp(-x_2)

    def inner_step(state, **kwargs):
        _pressure = kwargs.pop("pressure", pressure)

        R, P, M, F = state.position, state.momentum, state.mass, state.force
        R_b, P_b, M_b = state.box_position, state.box_momentum, state.box_mass

        N, dim = R.shape

        vol, box_fn = _npt_box_info(state)

        alpha = 1 + 1 / N
        G_e = box_force(alpha, vol, box_fn, R, P, M, F, _pressure, **kwargs)
        P_b = P_b + dt_2 * G_e
        P = exp_iL2(alpha, P, F, P_b / M_b)

        R_b = R_b + P_b / M_b * dt
        state = state.set(box_position=R_b)

        vol, box_fn = _npt_box_info(state)

        box = box_fn(vol)
        R = exp_iL1(box, R, P / M, P_b / M_b)
        F = force_fn(R, box=box, **kwargs)

        P = exp_iL2(alpha, P, F, P_b / M_b)
        G_e = box_force(alpha, vol, box_fn, R, P, M, F, _pressure, **kwargs)
        P_b = P_b + dt_2 * G_e

        return state.set(
            position=R,
            momentum=P,
            mass=M,
            force=F,
            box_position=R_b,
            box_momentum=P_b,
            box_mass=M_b,
        )

    def apply_fn(state, **kwargs):
        S = state
        if "kT" not in kwargs:
            _kT = kT
        else:
            _kT = kwargs.pop("kT")

        bc = barostat.update_mass(S.barostat, _kT)
        tc = thermostat.update_mass(S.thermostat, _kT)
        S = update_box_mass(S, _kT)

        P_b, bc = barostat.half_step(S.box_momentum, bc, _kT)
        P, tc = thermostat.half_step(S.momentum, tc, _kT)

        S = S.set(momentum=P, box_momentum=P_b)
        S = inner_step(S, **kwargs)

        KE = quantity.kinetic_energy(momentum=S.momentum, mass=S.mass)
        tc = tc.set(kinetic_energy=KE)

        KE_box = quantity.kinetic_energy(momentum=S.box_momentum, mass=S.box_mass)
        bc = bc.set(kinetic_energy=KE_box)

        P, tc = thermostat.half_step(S.momentum, tc, _kT)
        P_b, bc = barostat.half_step(S.box_momentum, bc, _kT)

        S = S.set(thermostat=tc, barostat=bc, momentum=P, box_momentum=P_b)

        return S

    return init_fn, apply_fn


def npt_nose_hoover_invariant(
    energy_fn: Callable[..., Array],
    state: NPTNoseHooverState,
    pressure: float,
    kT: float,
    **kwargs,
) -> float:
    """The conserved quantity for the NPT ensemble with a Nose-Hoover thermostat.

    This function is normally used for debugging the NPT simulation.

    Arguments:
      energy_fn: The energy function of the system.
      state: The current state of the system.
      pressure: The current goal pressure of the system.
      kT: The current goal temperature of the system.

    Returns:
      The Hamiltonian of the extended NPT dynamics.
    """
    volume, box_fn = _npt_box_info(state)
    PE = energy_fn(state.position, box=box_fn(volume), **kwargs)
    KE = kinetic_energy(state)

    DOF = state.position.size
    E = PE + KE

    c = state.thermostat
    E += c.momentum[0] ** 2 / (2 * c.mass[0]) + DOF * kT * c.position[0]
    for r, p, m in zip(c.position[1:], c.momentum[1:], c.mass[1:]):
        E += p**2 / (2 * m) + kT * r

    c = state.barostat
    for r, p, m in zip(c.position, c.momentum, c.mass):
        E += p**2 / (2 * m) + kT * r

    E += pressure * volume
    E += state.box_momentum**2 / (2 * state.box_mass)

    return E


"""Stochastic Simulations

JAX MD includes integrators for stochastic simulations of Langevin dynamics and
Brownian motion for systems in the NVT ensemble with a solvent.
"""


@dataclasses.dataclass
class Normal:
    """A simple normal distribution."""

    mean: jnp.ndarray
    var: jnp.ndarray

    def sample(self, key):
        mu, sigma = self.mean, jnp.sqrt(self.var)
        return mu + sigma * random.normal(key, mu.shape, dtype=mu.dtype)

    def log_prob(self, x):
        return (
            -0.5 * jnp.log(2 * jnp.pi * self.var)
            - 1 / (2 * self.var) * (x - self.mean) ** 2
        )


@dataclasses.dataclass
class NVTLangevinState:
    """A struct containing state information for the Langevin thermostat.

    Attributes:
      position: The current position of the particles. An ndarray of floats with
        shape `[n, spatial_dimension]`.
      momentum: The momentum of particles. An ndarray of floats with shape
        `[n, spatial_dimension]`.
      force: The (non-stochastic) force on particles. An ndarray of floats with
        shape `[n, spatial_dimension]`.
      mass: The mass of particles. Will either be a float or an ndarray of floats
        with shape `[n]`.
      rng: The current state of the random number generator.
    """

    position: Array
    momentum: Array
    force: Array
    mass: Array
    rng: Array

    @property
    def velocity(self) -> Array:
        return self.momentum / self.mass


@dispatch_by_state
def stochastic_step(state: NVTLangevinState, dt: float, kT: float, gamma: float):
    """A single stochastic step (the `O` step)."""
    c1 = jnp.exp(-gamma * dt)
    c2 = jnp.sqrt(kT * (1 - c1**2))
    momentum_dist = Normal(c1 * state.momentum, c2**2 * state.mass)
    key, split = random.split(state.rng)
    return state.set(momentum=momentum_dist.sample(split), rng=key)


def nvt_langevin(
    energy_or_force_fn: Callable[..., Array],
    shift_fn: ShiftFn,
    dt: float,
    kT: float,
    gamma: float = 0.1,
    center_velocity: bool = True,
    **sim_kwargs,
) -> Simulator:
    """Simulation in the NVT ensemble using the BAOAB Langevin thermostat.

    Samples from the canonical ensemble in which the number of particles (N),
    the system volume (V), and the temperature (T) are held constant. Langevin
    dynamics are stochastic and it is supposed that the system is interacting
    with fictitious microscopic degrees of freedom. An example of this would be
    large particles in a solvent such as water. Thus, Langevin dynamics are a
    stochastic ODE described by a friction coefficient and noise of a given
    covariance.

    Our implementation follows the paper [#davidcheck] by Davidchack, Ouldridge,
    and Tretyakov.

    Args:
      energy_or_force: A function that produces either an energy or a force from
        a set of particle positions specified as an ndarray of shape
        `[n, spatial_dimension]`.
      shift_fn: A function that displaces positions, `R`, by an amount `dR`. Both
        `R` and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
      dt: Floating point number specifying the timescale (step size) of the
        simulation.
      kT: Floating point number specifying the temperature in units of Boltzmann
        constant. To update the temperature dynamically during a simulation one
        should pass `kT` as a keyword argument to the step function.
      gamma: A float specifying the friction coefficient between the particles
        and the solvent.
      center_velocity: A boolean specifying whether or not the center of mass
        position should be subtracted.
    Returns:
      See above.

    .. rubric:: References
    .. [#carlon] R. L. Davidchack, T. E. Ouldridge, and M. V. Tretyakov.
      "New Langevin and gradient thermostats for rigid body dynamics."
      The Journal of Chemical Physics 142, 144114 (2015)
    """
    force_fn = quantity.canonicalize_force(energy_or_force_fn)

    @jit
    def init_fn(key, R, mass=f32(1.0), **kwargs):
        _kT = kwargs.pop("kT", kT)
        key, split = random.split(key)
        force = force_fn(R, **kwargs)
        state = NVTLangevinState(R, None, force, mass, key)
        state = canonicalize_mass(state)
        return initialize_momenta(state, split, _kT)

    @jit
    def step_fn(state, **kwargs):
        _dt = kwargs.pop("dt", dt)
        _kT = kwargs.pop("kT", kT)
        dt_2 = _dt / 2

        state = momentum_step(state, dt_2)
        state = position_step(state, shift_fn, dt_2, **kwargs)
        state = stochastic_step(state, _dt, _kT, gamma)
        state = position_step(state, shift_fn, dt_2, **kwargs)
        state = state.set(force=force_fn(state.position, **kwargs))
        state = momentum_step(state, dt_2)

        return state

    return init_fn, step_fn


@dataclasses.dataclass
class BrownianState:
    """A tuple containing state information for Brownian dynamics.

    Attributes:
      position: The current position of the particles. An ndarray of floats with
        shape `[n, spatial_dimension]`.
      mass: The mass of particles. Will either be a float or an ndarray of floats
        with shape `[n]`.
      rng: The current state of the random number generator.
    """

    position: Array
    mass: Array
    rng: Array


def brownian(
    energy_or_force: Callable[..., Array],
    shift: ShiftFn,
    dt: float,
    kT: float,
    gamma: float = 0.1,
) -> Simulator:
    """Simulation of Brownian dynamics.

    Simulates Brownian dynamics which are synonymous with the overdamped
    regime of Langevin dynamics. However, in this case we don't need to take into
    account velocity information and the dynamics simplify. Consequently, when
    Brownian dynamics can be used they will be faster than Langevin. As in the
    case of Langevin dynamics our implementation follows Carlon et al. [#carlon]_

    Args:
      energy_or_force: A function that produces either an energy or a force from
        a set of particle positions specified as an ndarray of shape
        `[n, spatial_dimension]`.
      shift_fn: A function that displaces positions, `R`, by an amount `dR`.
        Both `R` and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
      dt: Floating point number specifying the timescale (step size) of the
        simulation.
      kT: Floating point number specifying the temperature in units of Boltzmann
        constant. To update the temperature dynamically during a simulation one
        should pass `kT` as a keyword argument to the step function.
      gamma: A float specifying the friction coefficient between the particles
        and the solvent.

    Returns:
      See above.
    """

    force_fn = quantity.canonicalize_force(energy_or_force)

    dt, gamma = static_cast(dt, gamma)

    def init_fn(key, R, mass=f32(1)):
        state = BrownianState(R, mass, key)
        return canonicalize_mass(state)

    def apply_fn(state, **kwargs):
        _kT = kT if "kT" not in kwargs else kwargs["kT"]

        R, mass, key = dataclasses.astuple(state)

        key, split = random.split(key)

        F = force_fn(R, **kwargs)
        xi = random.normal(split, R.shape, R.dtype)

        nu = f32(1) / (mass * gamma)

        dR = F * dt * nu + jnp.sqrt(f32(2) * _kT * dt * nu) * xi
        R = shift(R, dR, **kwargs)

        return BrownianState(R, mass, key)  # pytype: disable=wrong-arg-count

    return init_fn, apply_fn


"""Experimental Simulations.


Below are simulation environments whose implementation is somewhat
experimental / preliminary. These environments might not be as ergonomic
as the more polished environments above.
"""


@dataclasses.dataclass
class SwapMCState:
    """A struct containing state information about a Hybrid Swap MC simulation.

    Attributes:
      md: A NVTNoseHooverState containing continuous molecular dynamics data.
      sigma: An `[n,]` array of particle radii.
      key: A JAX PRGNKey used for random number generation.
      neighbor: A NeighborList for the system.
    """

    md: NVTNoseHooverState
    sigma: Array
    key: Array
    neighbor: partition.NeighborList


# pytype: disable=wrong-arg-count
# pytype: disable=wrong-keyword-args
def hybrid_swap_mc(
    space_fns: space.Space,
    energy_fn: Callable[[Array, Array], Array],
    neighbor_fn: partition.NeighborFn,
    dt: float,
    kT: float,
    t_md: float,
    N_swap: int,
    sigma_fn: Optional[Callable[[Array], Array]] = None,
) -> Simulator:
    """Simulation of Hybrid Swap Monte-Carlo.

    This code simulates the hybrid Swap Monte Carlo algorithm introduced in
    Berthier et al. [#berthier]_
    Here an NVT simulation is performed for `t_md` time and then `N_swap` MC
    moves are performed that swap the radii of randomly chosen particles. The
    random swaps are accepted with Metropolis-Hastings step. Each call to the
    step function runs molecular dynamics for `t_md` and then performs the swaps.

    Note that this code doesn't feature some of the convenience functions in the
    other simulations. In particular, there is no support for dynamics keyword
    arguments and the energy function must be a simple callable of two variables:
    the distance between adjacent particles and the diameter of the particles.
    If you want support for a better notion of potential or dynamic keyword
    arguments, please file an issue!

    Args:
      space_fns: A tuple of a displacement function and a shift function defined
        in `space.py`.
      energy_fn: A function that computes the energy between one pair of
        particles as a function of the distance between the particles and the
        diameter. This function should not have been passed to `smap.xxx`.
      neighbor_fn: A function to construct neighbor lists outlined in
        `partition.py`.
      dt: The timestep used for the continuous time MD portion of the simulation.
      kT: The temperature of heat bath that the system is coupled to during MD.
      t_md: The time of each MD block.
      N_swap: The number of swapping moves between MD blocks.
      sigma_fn: An optional function for combining radii if they are to be
        non-additive.

    Returns:
      See above.

    .. rubric:: References
    .. [#berthier] L. Berthier, E. Flenner, C. J. Fullerton, C. Scalliet, and M. Singh.
      "Efficient swap algorithms for molecular dynamics simulations of
      equilibrium supercooled liquids", J. Stat. Mech. (2019) 064004
    """
    displacement_fn, shift_fn = space_fns
    metric_fn = space.metric(displacement_fn)
    nbr_metric_fn = space.map_neighbor(metric_fn)

    md_steps = int(t_md // dt)

    # Canonicalize the argument names to be dr and sigma.
    wrapped_energy_fn = lambda dr, sigma: energy_fn(dr, sigma)
    if sigma_fn is None:
        sigma_fn = lambda si, sj: 0.5 * (si + sj)
    nbr_energy_fn = smap.pair_neighbor_list(wrapped_energy_fn, metric_fn, sigma=sigma_fn)

    nvt_init_fn, nvt_step_fn = nvt_nose_hoover(
        nbr_energy_fn, shift_fn, dt, kT=kT, chain_length=3
    )

    def init_fn(key, position, sigma, nbrs=None):
        key, sim_key = random.split(key)
        nbrs = neighbor_fn(position, nbrs)  # pytype: disable=wrong-arg-count
        md_state = nvt_init_fn(sim_key, position, neighbor=nbrs, sigma=sigma)
        return SwapMCState(md_state, sigma, key, nbrs)  # pytype: disable=wrong-arg-count

    def md_step_fn(i, state):
        md, sigma, key, nbrs = dataclasses.unpack(state)
        md = nvt_step_fn(
            md, neighbor=nbrs, sigma=sigma
        )  # pytype: disable=wrong-keyword-args
        nbrs = neighbor_fn(md.position, nbrs)
        return SwapMCState(md, sigma, key, nbrs)  # pytype: disable=wrong-arg-count

    def swap_step_fn(i, state):
        md, sigma, key, nbrs = dataclasses.unpack(state)

        N = md.position.shape[0]

        # Swap a random pair of particle radii.
        key, particle_key, accept_key = random.split(key, 3)
        ij = random.randint(particle_key, (2,), jnp.array(0), jnp.array(N))
        new_sigma = sigma.at[ij].set([sigma[ij[1]], sigma[ij[0]]])

        # Collect neighborhoods around the two swapped particles.
        nbrs_ij = nbrs.idx[ij]
        R_ij = md.position[ij]
        R_neigh = md.position[nbrs_ij]

        sigma_ij = sigma[ij][:, None]
        sigma_neigh = sigma[nbrs_ij]

        new_sigma_ij = new_sigma[ij][:, None]
        new_sigma_neigh = new_sigma[nbrs_ij]

        dR = nbr_metric_fn(R_ij, R_neigh)

        # Compute the energy before the swap.
        energy = energy_fn(dR, sigma_fn(sigma_ij, sigma_neigh))
        energy = jnp.sum(energy * (nbrs_ij < N))

        # Compute the energy after the swap.
        new_energy = energy_fn(dR, sigma_fn(new_sigma_ij, new_sigma_neigh))
        new_energy = jnp.sum(new_energy * (nbrs_ij < N))

        # Accept or reject with a metropolis probability.
        p = random.uniform(accept_key, ())
        accept_prob = jnp.minimum(1, jnp.exp(-(new_energy - energy) / kT))
        sigma = jnp.where(p < accept_prob, new_sigma, sigma)

        return SwapMCState(md, sigma, key, nbrs)  # pytype: disable=wrong-arg-count

    def block_fn(state):
        state = lax.fori_loop(0, md_steps, md_step_fn, state)
        state = lax.fori_loop(0, N_swap, swap_step_fn, state)
        return state

    return init_fn, block_fn


# pytype: enable=wrong-arg-count
# pytype: enable=wrong-keyword-args


def temp_rescale(
    energy_or_force_fn: Callable[..., Array],
    shift_fn: ShiftFn,
    dt: float,
    kT: float,
    window: float,
    fraction: float,
    **sim_kwargs,
) -> Simulator:
    """Simulation using explicit velocity rescaling.

    Rescale the velocities of atoms explicitly so that the desired temperature is
    reached.

    Args:
      energy_or_force: A function that produces either an energy or a force from
        a set of particle positions specified as an ndarray of shape
        `[n, spatial_dimension]`.
      shift_fn: A function that displaces positions, `R`, by an amount `dR`.
        Both `R` and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
      dt: Floating point number specifying the timescale (step size) of the
        simulation.
      kT: Floating point number specifying the temperature in units of Boltzmann
        constant. To update the temperature dynamically during a simulation one
        should pass `kT` as a keyword argument to the step function.
      window: Floating point number specifying the temperature window outside which
        rescaling is performed. Measured in units of `kT`.
      fraction: Floating point number which determines the amount of rescaling
        applied to the velocities. Takes values from 0.0-1.0.
    Returns:
      See above.

    .. rubric:: References
    .. [#berendsen84] Woodcock, L. V.
      "ISOTHERMAL MOLECULAR DYNAMICS CALCULATIONS FOR LIQUID SALTS."
      Chem. Phys. Lett. 1971, 10, 257–261.
    """
    force_fn = quantity.canonicalize_force(energy_or_force_fn)
    dt = f32(dt)

    def velocity_rescale(state, window, fraction, kT):
        """Rescale the momentum if the the difference between current and target
        temperature is more than the window"""
        kT_current = temperature(state)
        cond = jnp.abs(kT_current - kT) > window
        kT_target = kT_current - fraction * (kT_current - kT)
        lam = jnp.where(cond, jnp.sqrt(kT_target / kT_current), 1)
        new_momentum = tree_util.tree_map(lambda p: p * lam, state.momentum)
        return state.set(momentum=new_momentum)

    def init_fn(key, R, mass=f32(1.0), **kwargs):
        # Reuse the NVEState dataclass
        state = NVEState(R, None, force_fn(R, **kwargs), mass)
        state = canonicalize_mass(state)
        return initialize_momenta(state, key, kT)

    def apply_fn(state, **kwargs):
        state = velocity_rescale(state, window, fraction, kT)
        state = velocity_verlet(force_fn, shift_fn, dt, state, **kwargs)
        return state

    return init_fn, apply_fn


def temp_berendsen(
    energy_or_force_fn: Callable[..., Array],
    shift_fn: ShiftFn,
    dt: float,
    kT: float,
    tau: float,
    **sim_kwargs,
) -> Simulator:
    """Simulation using the Berendsen thermostat.

    Berendsen (weak coupling) thermostat rescales the velocities of atoms such
    that the desired temperature is reached. This rescaling is performed at each
    timestep (dt) and the rescaling factor is calculated using
    Eq.10 Berendsen et al. [#berendsen84]_.

    Args:
      energy_or_force: A function that produces either an energy or a force from
        a set of particle positions specified as an ndarray of shape
        `[n, spatial_dimension]`.
      shift_fn: A function that displaces positions, `R`, by an amount `dR`.
        Both `R` and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
      dt: Floating point number specifying the timescale (step size) of the
        simulation.
      kT: Floating point number specifying the temperature in units of Boltzmann
        constant. To update the temperature dynamically during a simulation one
        should pass `kT` as a keyword argument to the step function.
      tau: A floating point number determining how fast the temperature
        is relaxed during the simulation. Measured in units of `dt`.
    Returns:
      See above.

    .. rubric:: References
    .. [#berendsen84] H. J. C. Berendsen, J. P. M. Postma, W. F. van Gunsteren, A. DiNola, J. R. Haak.
      "Molecular dynamics with coupling to an external bath."
      J. Chem. Phys. 15 October 1984; 81 (8): 3684-3690.
    """
    force_fn = quantity.canonicalize_force(energy_or_force_fn)
    dt = f32(dt)

    def berendsen_update(state, tau, kT, dt):
        """Rescaling the momentum of the particle by the factor lam."""
        _kT = temperature(state)
        lam = jnp.sqrt(1 + ((dt / tau) * ((kT / _kT) - 1)))
        new_momentum = tree_util.tree_map(lambda p: p * lam, state.momentum)
        return state.set(momentum=new_momentum)

    def init_fn(key, R, mass=f32(1.0), **kwargs):
        # Reuse the NVEState dataclass
        state = NVEState(R, None, force_fn(R, **kwargs), mass)
        state = canonicalize_mass(state)
        return initialize_momenta(state, key, kT)

    def apply_fn(state, **kwargs):
        state = berendsen_update(state, tau, kT, dt)
        state = velocity_verlet(force_fn, shift_fn, dt, state, **kwargs)
        return state

    return init_fn, apply_fn


def nvk(
    energy_or_force_fn: Callable[..., Array],
    shift_fn: ShiftFn,
    dt: float,
    kT: float,
    **sim_kwargs,
) -> Simulator:
    """Simulation in the NVK (isokinetic) ensemble using the Gaussian thermostat.

    Samples from the isokinetic ensemble in which the number of particles (N),
    the system volume (V), and the kinetic energy (K) are held constant. A
    Gaussian thermostat is used for the integration and the kinetic energy is
    held constant during the simulation. The implementation follows the steps
    described in [#minary2003]_ and [#zhang97]_. See section 4(B) equation
    4.12-4.17 in [#minary2003]_ for detailed description.

    Args:
      energy_or_force: A function that produces either an energy or a force from
        a set of particle positions specified as an ndarray of shape
        `[n, spatial_dimension]`.
      shift_fn: A function that displaces positions, `R`, by an amount `dR`.
        Both `R` and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
      dt: Floating point number specifying the timescale (step size) of the
        simulation.
      kT: Floating point number specifying the temperature in units of Boltzmann
        constant. To update the temperature dynamically during a simulation one
        should pass `kT` as a keyword argument to the step function.
    Returns:
      See above.

    .. rubric:: References
    .. [#minary2003] Minary, Peter and Martyna, Glenn J. and Tuckerman, Mark E.
      "Algorithms and novel applications based on the isokinetic ensemble. I.
      Biophysical and path integral molecular dynamics"
      J. Chem. Phys., Vol. 118, No. 6, 8 February 2003.
    .. [#zhang97] Zhang, Fei.
      "Operator-splitting integrators for constant-temperature molecular dynamics"
      J. Chem. Phys. 106, 6102–6106 (1997).
    """
    force_fn = quantity.canonicalize_force(energy_or_force_fn)
    dt = f32(dt)
    dt_2 = f32(dt / 2)

    def momentum_update(state, KE):
        # eps to avoid edge cases when forces are zero
        eps = 1e-16

        # Equation 4.13 to compute a and b
        update_fn = lambda f, p, m: f * p / m
        a = (
            util.high_precision_sum(update_fn(state.force, state.momentum, state.mass))
            + eps
        )
        b = util.high_precision_sum(update_fn(state.force, state.force, state.mass)) + eps
        a /= 2.0 * KE
        b /= 2.0 * KE

        # Equation 4.12 to compute s(t) and s_dot(t)
        b_sqrt = jnp.sqrt(b)
        s_t = ((a / b) * (jnp.cosh(dt_2 * b_sqrt) - 1.0)) + jnp.sinh(
            dt_2 * b_sqrt
        ) / b_sqrt
        s_dot_t = (b_sqrt * (a / b) * jnp.sinh(dt_2 * b_sqrt)) + jnp.cosh(dt_2 * b_sqrt)

        # Get the new momentum using Equation 4.15
        new_momentum = tree_util.tree_map(
            lambda p, f, s, sdot: (p + f * s) / sdot,
            state.momentum,
            state.force,
            s_t,
            s_dot_t,
        )
        return state.set(momentum=new_momentum)

    def position_update(state, shift_fn, **kwargs):
        if isinstance(shift_fn, Callable):
            shift_fn = tree_util.tree_map(lambda r: shift_fn, state.position)
        # Get the new positions using Equation 4.16 (Should read r = r + dt * p / m)
        new_position = tree_util.tree_map(
            lambda s_fn, r, v: s_fn(r, dt * v, **kwargs),
            shift_fn,
            state.position,
            state.velocity,
        )
        return state.set(position=new_position)

    def init_fn(key, R, mass=f32(1.0), **kwargs):
        _kT = kwargs.pop("kT", kT)
        key, split = random.split(key)
        # Reuse the NVEState dataclass
        state = NVEState(R, None, force_fn(R, **kwargs), mass)
        state = canonicalize_mass(state)
        return initialize_momenta(state, split, _kT)

    def apply_fn(state, **kwargs):
        _KE = kinetic_energy(state)
        state = momentum_update(state, _KE)
        state = position_update(state, shift_fn)
        state = state.set(force=force_fn(state.position, **kwargs))
        state = momentum_update(state, _KE)
        return state

    return init_fn, apply_fn


def temp_csvr(
    energy_or_force_fn: Callable[..., Array],
    shift_fn: ShiftFn,
    dt: float,
    kT: float,
    tau: float,
    **sim_kwargs,
) -> Simulator:
    """Simulation using the canonical sampling through velocity rescaling (CSVR) thermostat.

    Samples from the canonical ensemble in which the number of particles (N),
    the system volume (V), and the temperature (T) are held constant. CSVR
    algorithm samples the canonical distribution by rescaling the velocities
    by a appropriately chosen random factor. At each timestep (dt) the rescaling
    takes place and the rescaling factor is calculated using
    A7 Bussi et al. [#bussi2007]_. CSVR updates to the velocity are stochastic in
    nature and unlike the Berendsen thermostat it samples the true canonical
    distribution [#Braun2018]_.

    Args:
      energy_or_force: A function that produces either an energy or a force from
        a set of particle positions specified as an ndarray of shape
        `[n, spatial_dimension]`.
      shift_fn: A function that displaces positions, `R`, by an amount `dR`.
        Both `R` and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
      dt: Floating point number specifying the timescale (step size) of the
        simulation.
      kT: Floating point number specifying the temperature in units of Boltzmann
        constant. To update the temperature dynamically during a simulation one
        should pass `kT` as a keyword argument to the step function.
      tau: A floating point number determining how fast the temperature
        is relaxed during the simulation. Measured in units of `dt`.
    Returns:
      See above.

    .. rubric:: References
    .. [#bussi2007] Bussi G, Donadio D, Parrinello M.
      "Canonical sampling through velocity rescaling."
      The Journal of chemical physics, 126(1), 014101.
    .. [#Braun2018] Efrem Braun, Seyed Mohamad Moosavi, and Berend Smit.
      "Anomalous Effects of Velocity Rescaling Algorithms: The Flying Ice Cube Effect Revisited."
      Journal of Chemical Theory and Computation 2018 14 (10), 5262-5272.
    """
    force_fn = quantity.canonicalize_force(energy_or_force_fn)
    dt = f32(dt)

    def sum_noises(state, key):
        """Sum of N independent gaussian noises squared.
        Adapted from https://github.com/GiovanniBussi/StochasticVelocityRescaling
        For more details see Eq.A7 Bussi et al. [#bussi2007]_"""
        dof = quantity.count_dof(state.position) - 1
        _dtype = state.position.dtype

        if dof == 0:
            """If there are no terms return zero."""
            return 0

        elif dof == 1:
            """For a single noise term, directly calculate the square of the Gaussian
            noise value."""
            rr = random.normal(key, dtype=_dtype)
            return rr * rr

        elif dof % 2 == 0:
            """For an even number of noise terms, use the gamma-distributed random
            number generator"""
            return 2.0 * random.gamma(key, dof // 2, dtype=_dtype)

        else:
            """For an odd number of noise terms, sum two terms: one from the
            gamma-distributed generator and another from the square of a
            Gaussian-distributed random number."""
            rr = random.normal(key, dtype=_dtype)
            return 2.0 * random.gamma(key, (dof - 1) // 2, dtype=_dtype) + (rr * rr)

    def csvr_update(state, tau, kT, dt):
        """Update the momentum by an scaling factor as described by
        Eq.A7 Bussi et al. [#bussi2007]_"""
        key, split = random.split(state.rng)
        dof = quantity.count_dof(state.position)

        _kT = temperature(state)

        KE_old = dof * _kT / 2
        KE_new = dof * kT / 2

        r1 = random.normal(key, dtype=state.position.dtype)
        r2 = sum_noises(state, key)

        c1 = jnp.exp(-dt / tau)
        c2 = (1 - c1) * KE_new / KE_old / dof

        scale = c1 + (c2 * ((r1 * r1) + r2)) + (2 * r1 * jnp.sqrt(c1 * c2))
        lam = jnp.sqrt(scale)

        new_momentum = tree_util.tree_map(lambda p: p * lam, state.momentum)
        return state.set(momentum=new_momentum, rng=key)

    def init_fn(key, R, mass=f32(1.0), **kwargs):
        _kT = kwargs.pop("kT", kT)
        key, split = random.split(key)
        # Reuse the NVTLangevinState dataclass
        state = NVTLangevinState(R, None, force_fn(R, **kwargs), mass, key)
        state = canonicalize_mass(state)
        return initialize_momenta(state, split, _kT)

    def apply_fn(state, **kwargs):
        state = csvr_update(state, tau, kT, dt)
        state = velocity_verlet(force_fn, shift_fn, dt, state, **kwargs)
        return state

    return init_fn, apply_fn



import logging

import jax.numpy as jnp
import numpy as np
import optax
from flax import traverse_util
from flax.core.frozen_dict import freeze
from optax._src import base

from apax.optimizer.optimizers import ademamix, sam

log = logging.getLogger(__name__)


def cyclic_cosine_decay_schedule(
    init_value: float,
    steps_per_epoch,
    period: int,
    decay_factor: float = 0.9,
) -> base.Schedule:
    r"""Returns a function which implements cyclic cosine learning rate decay.

    Args:
        init_value: An initial value for the learning rate.

    Returns:
        schedule
        A function that maps step counts to values.
    """

    def schedule(count):
        cycle = count // (period * steps_per_epoch)
        step_in_period = jnp.mod(count, period * steps_per_epoch)
        arg = np.pi * step_in_period / (period * steps_per_epoch)
        lr = init_value / 2 * (jnp.cos(arg) + 1)
        lr = lr * (decay_factor**cycle)
        return lr

    return schedule


def get_schedule(
    lr: float,
    n_epochs: int,
    steps_per_epoch: int,
    schedule_kwargs: dict,
) -> optax._src.base.Schedule:
    """
    builds a linear learning rate schedule.
    """
    schedule_kwargs = schedule_kwargs.copy()
    name = schedule_kwargs.pop("name")
    if name == "linear":
        lr_schedule = optax.linear_schedule(
            init_value=lr, transition_steps=n_epochs * steps_per_epoch, **schedule_kwargs
        )
    elif name == "cyclic_cosine":
        lr_schedule = cyclic_cosine_decay_schedule(lr, steps_per_epoch, **schedule_kwargs)
    else:
        raise KeyError(f"unknown learning rate schedule: {name}")
    return lr_schedule


class OptimizerFactory:
    def __init__(
        self, opt, n_epochs, steps_per_epoch, gradient_clipping, kwargs, schedule
    ) -> None:
        self.opt = opt
        self.n_epochs = n_epochs
        self.steps_per_epoch = steps_per_epoch
        self.gradient_clipping = gradient_clipping
        self.kwargs = kwargs
        self.schedule = schedule

    def create(self, lr):
        if lr <= 1e-7:
            optimizer = optax.set_to_zero()
        else:
            schedule = get_schedule(
                lr, self.n_epochs, self.steps_per_epoch, self.schedule
            )
            optimizer = optax.chain(
                optax.clip(self.gradient_clipping),
                self.opt(schedule, **self.kwargs),
                optax.zero_nans(),
            )
        return optimizer


def get_opt(
    params,
    n_epochs: int,
    steps_per_epoch: int,
    emb_lr: float = 0.02,
    nn_lr: float = 0.03,
    scale_lr: float = 0.001,
    shift_lr: float = 0.05,
    zbl_lr: float = 0.001,
    rep_scale_lr: float = 0.001,
    rep_prefactor_lr: float = 0.0001,
    gradient_clipping=1000.0,
    name: str = "adam",
    kwargs: dict = {},
    schedule: dict = {},
) -> optax._src.base.GradientTransformation:
    """
    Builds an optimizer with different learning rates for each parameter group.
    Several `optax` optimizers are supported.
    """

    log.info("Initializing Optimizer")
    if name == "sam":
        opt = sam
    elif name == "ademamix":
        opt = ademamix
    else:
        opt = getattr(optax, name)

    opt_fac = OptimizerFactory(
        opt, n_epochs, steps_per_epoch, gradient_clipping, kwargs, schedule
    )

    nn_opt = opt_fac.create(nn_lr)
    emb_opt = opt_fac.create(emb_lr)
    scale_opt = opt_fac.create(scale_lr)
    shift_opt = opt_fac.create(shift_lr)
    zbl_opt = opt_fac.create(zbl_lr)
    rep_scale_opt = opt_fac.create(rep_scale_lr)
    rep_prefactor_opt = opt_fac.create(rep_prefactor_lr)

    partition_optimizers = {
        "w": nn_opt,
        "b": nn_opt,
        "atomic_type_embedding": emb_opt,
        "scale_per_element": scale_opt,
        "shift_per_element": shift_opt,
        "a_exp": zbl_opt,
        "a_num": zbl_opt,
        "coefficients": zbl_opt,
        "exponents": zbl_opt,
        "rep_scale": rep_scale_opt,
        "rep_prefactor": rep_prefactor_opt,
        "kernel": nn_opt,
        "bias": nn_opt,
        "embedding": emb_opt,
        "weights_K": nn_opt,
        "weights_Q": nn_opt,
        "weights_V": nn_opt,
        "scale": scale_opt,
    }

    param_partitions = freeze(
        traverse_util.path_aware_map(lambda path, v: path[-1], params)
    )
    tx = optax.multi_transform(partition_optimizers, param_partitions)

    return tx



from typing import Callable, Literal, Union

import jax.numpy as jnp
from pydantic import BaseModel, TypeAdapter


class ConstraintBase(BaseModel):
    """Base class for constraints.
    Constraints work by implementing a create method.
    This method accepts a reference state which to
    compare to subsequent ones and returns a callable which applies
    the constraint during simulations.
    """

    def create(self, ref_stat, system) -> Callable:
        pass


class FixAtoms(ConstraintBase, extra="forbid"):
    """ """

    name: Literal["fixatoms"] = "fixatoms"
    indices: list[int]

    def create(self, ref_state, system) -> Callable:
        indices = jnp.array(self.indices, dtype=jnp.int64)

        ref_position = ref_state.position[indices]

        def fn(state):
            position = state.position
            force = state.force
            zero_force = jnp.zeros_like(ref_position)
            position = position.at[indices].set(ref_position)
            force = force.at[indices].set(zero_force)
            state = state.set(position=position, force=force)
            return state

        return fn


class FixLayer(ConstraintBase, extra="forbid"):
    """ """

    name: Literal["fixlayer"] = "fixlayer"
    upper_limit: float
    lower_limit: float

    def create(self, ref_state, system) -> Callable:
        pos = ref_state.position @ system.box
        z_coordinates = pos[:, 2]

        indices = jnp.where(
            (self.lower_limit <= z_coordinates) & (z_coordinates <= self.upper_limit)
        )

        ref_position = ref_state.position[indices]

        def fn(state):
            position = state.position
            force = state.force
            zero_force = jnp.zeros_like(ref_position)
            position = position.at[indices].set(ref_position)
            force = force.at[indices].set(zero_force)
            state = state.set(position=position, force=force)
            return state

        return fn


Constraint = TypeAdapter(Union[FixAtoms, FixLayer]).validate_python



from dataclasses import field
from typing import Any, Callable, List

import flax.linen as nn
import jax.numpy as jnp

from apax.layers.activation import swish
from apax.layers.ntk_linear import NTKLinear
from apax.utils.convert import str_to_dtype


class AtomisticReadout(nn.Module):
    units: List[int] = field(default_factory=lambda: [32, 32])
    activation_fn: Callable = swish
    w_init: str = "normal"
    b_init: str = "zeros"
    use_ntk: bool = True
    n_shallow_ensemble: int = 0
    is_feature_fn: bool = False
    dtype: Any = jnp.float32

    def setup(self):
        units = list(self.units)
        if not self.is_feature_fn:
            readout_unit = [1]
            if self.n_shallow_ensemble > 0:
                readout_unit = [self.n_shallow_ensemble]
            units += readout_unit

        dtype = str_to_dtype(self.dtype)

        dense = []
        for ii, n_hidden in enumerate(units):
            layer = NTKLinear(
                n_hidden,
                w_init=self.w_init,
                b_init=self.b_init,
                use_ntk=self.use_ntk,
                dtype=dtype,
                name=f"dense_{ii}",
            )
            dense.append(layer)
            if ii < len(units) - 1:
                dense.append(swish)
        self.sequential = nn.Sequential(dense, name="readout")

    def __call__(self, x):
        h = self.sequential(x)
        # TODO should we move aggregation here?
        return h



from typing import Any

import einops
import flax.linen as nn
import jax.numpy as jnp
import numpy as np

from apax.layers.initializers import uniform_range
from apax.utils.convert import str_to_dtype


class GaussianBasis(nn.Module):
    n_basis: int = 7
    r_min: float = 0.5
    r_max: float = 6.0
    dtype: Any = jnp.float32

    def setup(self):
        dtype = str_to_dtype(self.dtype)

        self.betta = self.n_basis**2 / self.r_max**2
        self.rad_norm = (2.0 * self.betta / np.pi) ** 0.25
        shifts = self.r_min + (self.r_max - self.r_min) / self.n_basis * np.arange(
            self.n_basis
        )

        # shape: 1 x n_basis
        shifts = einops.repeat(shifts, "n_basis -> 1 n_basis")
        self.shifts = jnp.asarray(shifts, dtype=dtype)

    def __call__(self, dr):
        dr = einops.repeat(dr, "neighbors -> neighbors 1")
        # 1 x n_basis, neighbors x 1 -> neighbors x n_basis
        distances = self.shifts - dr

        # shape: neighbors x n_basis
        basis = jnp.exp(-self.betta * (distances**2))
        basis = self.rad_norm * basis

        return basis


class BesselBasis(nn.Module):
    """Non-orthogonalized basis functions of Kocer
    https://doi.org/10.1063/1.5086167
    """

    n_basis: int = 7
    r_max: float = 6.0
    dtype: Any = jnp.float32

    def setup(self):
        dtype = str_to_dtype(self.dtype)
        self.n = jnp.arange(self.n_basis, dtype=dtype)

    def __call__(self, dr):
        dr = einops.repeat(dr, "neighbors -> neighbors 1")
        a = (-1) ** self.n * (jnp.sqrt(2) * np.pi / (self.r_max ** (3 / 2)))
        b = (self.n + 1) * (self.n + 2) / jnp.sqrt((self.n + 1) ** 2 * (self.n + 2) ** 2)
        s1 = jnp.sinc((self.n + 1) * dr / self.r_max)
        s2 = jnp.sinc((self.n + 2) * dr / self.r_max)
        basis = a * b * (s1 + s2)
        return basis


class RadialFunction(nn.Module):
    n_radial: int = 5
    basis_fn: nn.Module = GaussianBasis()
    n_species: int = 119
    emb_init: str = "uniform"
    use_embed_norm: bool = True
    one_sided_dist: bool = False
    dtype: Any = jnp.float32

    def setup(self):
        dtype = str_to_dtype(self.dtype)
        self.r_max = self.basis_fn.r_max
        self.embed_norm = jnp.array(1.0 / np.sqrt(self.basis_fn.n_basis), dtype=dtype)
        if self.one_sided_dist:
            lower_bound = 0.0
        else:
            lower_bound = -1.0

        if self.emb_init is not None:
            self._n_radial = self.n_radial
            if self.emb_init == "uniform":
                emb_initializer = uniform_range(lower_bound, 1.0, dtype=dtype)
                self.embeddings = self.param(
                    "atomic_type_embedding",
                    emb_initializer,
                    (
                        self.n_species,
                        self.n_species,
                        self.n_radial,
                        self.basis_fn.n_basis,
                    ),
                    dtype,
                )
            else:
                raise ValueError(
                    "Currently only uniformly initialized embeddings and no embeddings"
                    " are implemented."
                )
        else:
            self._n_radial = self.basis_fn.n_basis

    def __call__(self, dr, Z_i, Z_j):
        dtype = str_to_dtype(self.dtype)
        dr = dr.astype(dtype)
        # basis shape: neighbors x n_basis
        basis = self.basis_fn(dr)

        if self.emb_init is None:
            radial_function = basis
        else:
            # coeffs shape: n_neighbors x n_radialx n_basis
            species_pair_coeffs = self.embeddings[
                Z_j, Z_i, ...
            ]  # reverse convention to match original
            if self.use_embed_norm:
                species_pair_coeffs = self.embed_norm * species_pair_coeffs

            # radial shape: neighbors x n_radial
            radial_function = einops.einsum(
                species_pair_coeffs, basis, "nbrs radial basis, nbrs basis -> nbrs radial"
            )

        # shape: neighbors
        dr_clipped = jnp.clip(dr, a_max=self.r_max)
        cos_cutoff = 0.5 * (jnp.cos(np.pi * dr_clipped / self.r_max) + 1.0)
        cutoff = einops.repeat(cos_cutoff, "neighbors -> neighbors 1")

        radial_function = radial_function * cutoff

        assert radial_function.dtype == dtype

        return radial_function



import jax.numpy as jnp

# import numpy as np

# def tril_2d_indices(n: int):
#     indices = np.zeros((int(n * (n + 1) / 2), 2), dtype=int)
#     sparse_idx = 0
#     for i in range(0, n):
#         for j in range(i, n):
#             indices[sparse_idx] = i, j
#             sparse_idx += 1

#     return jnp.asarray(indices)


# def tril_3d_indices(n: int):
#     indices = np.zeros((int(n * (n + 1) * (n + 2) / 6), 3), dtype=int)
#     sparse_idx = 0
#     for i in range(0, n):
#         for j in range(i, n):
#             for k in range(j, n):
#                 indices[sparse_idx] = i, j, k
#                 sparse_idx += 1

#     return jnp.asarray(indices)


def tril_2d_indices(n_radial):
    tril_idxs = []
    for i in range(n_radial):
        tril_idxs.append([i, i])
        for j in range(i + 1, n_radial):
            tril_idxs.append([i, j])
    tril_idxs = jnp.array(tril_idxs)
    return tril_idxs


def tril_3d_indices(n_radial):
    tril_idxs = []
    for i in range(n_radial):
        tril_idxs.append([i, i, i])
        for j in range(n_radial):
            if j != i:
                tril_idxs.append([i, j, j])
        for j in range(i + 1, n_radial):
            for k in range(j + 1, n_radial):
                tril_idxs.append([i, j, k])
    tril_idxs = jnp.array(tril_idxs)
    return tril_idxs



import os
import warnings

import jax

os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
jax.config.update("jax_enable_x64", True)
from apax.utils.helpers import setup_ase

setup_ase()

warnings.filterwarnings("ignore", message=".*os.fork()*")



from dataclasses import field
from typing import Any

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from ase import data
from jax import vmap

from apax.layers.masking import mask_by_atom, mask_by_neighbor
from apax.utils.convert import str_to_dtype
from apax.utils.jax_md_reduced import space
from apax.utils.math import fp64_sum


def inverse_softplus(x):
    return jnp.log(jnp.exp(x) - 1.0)


class EmpiricalEnergyTerm(nn.Module):
    dtype: Any = jnp.float32


class ZBLRepulsion(EmpiricalEnergyTerm):
    r_max: float = 2.0
    apply_mask: bool = True

    def setup(self):
        self.distance = vmap(space.distance, 0, 0)

        coeffs = jnp.array([0.18175, 0.50986, 0.28022, 0.02817])[:, None]
        coeffs_isp = inverse_softplus(coeffs)
        rep_scale_isp = inverse_softplus(0.1)

        self.a_exp = 0.23
        self.a_num = 0.46850
        self.coefficients = self.param(
            "coefficients",
            nn.initializers.constant(coeffs_isp),
            (4, 1),
        )

        self.exponents = jnp.array([3.19980, 0.94229, 0.4029, 0.20162])[:, None]

        self.rep_scale = self.param(
            "rep_scale", nn.initializers.constant(rep_scale_isp), (1,)
        )

    def __call__(self, R, dr_vec, Z, idx, box, properties):
        dtype = str_to_dtype(self.dtype)
        # Z shape n_atoms

        idx_i, idx_j = idx[0], idx[1]

        # shape: neighbors
        Z_i, Z_j = Z[idx_i, ...], Z[idx_j, ...]

        # dr shape: neighbors
        dr = self.distance(dr_vec).astype(dtype)

        dr = jnp.clip(dr, a_min=0.02, a_max=self.r_max)
        cos_cutoff = 0.5 * (jnp.cos(np.pi * dr / self.r_max) + 1.0)

        # Ensure positive parameters
        a_exp = self.a_exp
        a_num = self.a_num
        coefficients = jax.nn.softplus(self.coefficients)
        exponents = self.exponents
        rep_scale = jax.nn.softplus(self.rep_scale)[0]

        a_divisor = Z_i**a_exp + Z_j**a_exp
        dist = dr * a_divisor / a_num
        f = coefficients * jnp.exp(-exponents * dist)
        f = jnp.sum(f, axis=0)

        E_ij = Z_i * Z_j / dr * f * cos_cutoff
        if self.apply_mask:
            E_ij = mask_by_neighbor(E_ij, idx)
        E = 0.5 * rep_scale * fp64_sum(E_ij)
        return E


class ExponentialRepulsion(EmpiricalEnergyTerm):
    r_max: float = 2.0
    apply_mask: bool = True

    def setup(self):
        self.distance = vmap(space.distance, 0, 0)

        radii = data.covalent_radii * 0.8
        self.rscale = self.param("rep_scale", nn.initializers.constant(radii), (119,))

        self.prefactor = self.param(
            "rep_prefactor", nn.initializers.constant(100.0), (119,)
        )

    def __call__(self, R, dr_vec, Z, idx, box, properties):
        dtype = str_to_dtype(self.dtype)

        # Z shape n_atoms
        idx_i, idx_j = idx[0], idx[1]

        # shape: neighbors
        Z_i, Z_j = Z[idx_i, ...], Z[idx_j, ...]

        # dr shape: neighbors
        dr = self.distance(dr_vec).astype(dtype)

        dr = jnp.clip(dr, a_min=0.02, a_max=self.r_max)
        cos_cutoff = 0.5 * (jnp.cos(np.pi * dr / self.r_max) + 1.0)

        # Ensure positive parameters
        A_i, A_j = (
            jax.numpy.abs(self.prefactor[Z_i]),
            jax.numpy.abs(self.prefactor[Z_j]),
        )
        R_i, R_j = jax.numpy.abs(self.rscale[Z_i]), jax.numpy.abs(self.rscale[Z_j])

        f = A_i * A_j * jnp.exp(-dr * (R_i + R_j) / (R_i * R_j)) / dr**2

        E_ij = f * cos_cutoff
        if self.apply_mask:
            E_ij = mask_by_neighbor(E_ij, idx)
        E = fp64_sum(E_ij)
        return E


class LatentEwald(EmpiricalEnergyTerm):
    """Latent Ewald summation by Cheng https://arxiv.org/abs/2408.15165
    Requires a property head which predicts 'charge' per atom.
    """

    kgrid: list[int] = field(default_factory=lambda: [2, 2, 2])
    sigma: float = 1.0
    apply_mask: bool = True

    def __call__(self, R, dr_vec, Z, idx, box, properties):
        # Z shape n_atoms
        if "charge" not in properties:
            raise KeyError(
                "property 'charge' not found. Make sure to predict it in the model section"
            )

        q = properties["charge"]

        V = jnp.linalg.det(box)
        Lx, Ly, Lz = jnp.linalg.norm(box, axis=1)

        k_range_x = 2 * np.pi * jnp.arange(1, self.kgrid[0]) / Lx
        k_range_y = 2 * np.pi * jnp.arange(1, self.kgrid[1]) / Ly
        k_range_z = 2 * np.pi * jnp.arange(1, self.kgrid[2]) / Lz

        kx, ky, kz = jnp.meshgrid(k_range_x, k_range_y, k_range_z)
        k = jnp.reshape(jnp.stack((kx, ky, kz), axis=-1), (-1, 3))

        k2 = jnp.sum(k**2, axis=-1)

        sf_k = q * jnp.exp(1j * jnp.einsum("id,jd->ij", R, k))
        if self.apply_mask:
            sf_k = mask_by_atom(sf_k, Z)
        sf = jnp.sum(sf_k, axis=0)
        S2 = jnp.abs(sf) ** 2

        E_lr = -fp64_sum(jnp.exp(-k2 * (self.sigma**2) / 2) / k2 * S2) / V

        return E_lr


all_corrections = {
    "zbl": ZBLRepulsion,
    "exponential": ExponentialRepulsion,
    "latent_ewald": LatentEwald,
}



import functools
import logging
import time
from functools import partial
from typing import Callable, Optional

import jax
import jax.numpy as jnp
import numpy as np
from clu import metrics
from flax.training.train_state import TrainState
from jax import tree_util
from jax.experimental import mesh_utils
from jax.sharding import PositionalSharding
from tqdm import trange

from apax.data.input_pipeline import InMemoryDataset
from apax.train.checkpoints import CheckpointManager, load_state
from apax.train.parameters import EMAParameters

log = logging.getLogger(__name__)


def fit(
    state: TrainState,
    train_ds: InMemoryDataset,
    loss_fn,
    Metrics: metrics.Collection,
    callbacks: list,
    n_epochs: int,
    ckpt_dir,
    ckpt_interval: int = 1,
    val_ds: Optional[InMemoryDataset] = None,
    patience: Optional[int] = None,
    disable_pbar: bool = False,
    disable_batch_pbar: bool = True,
    is_ensemble=False,
    data_parallel=True,
    ema_handler: Optional[EMAParameters] = None,
):
    """
    Trains the model using the provided training dataset.

    Parameters
    ----------
    state :
        The initial state of the model.
    train_ds : InMemoryDataset
        The training dataset.
    loss_fn :
        The loss function to be minimized.
    Metrics metrics.Collection :
        Collection of metrics to evaluate during training.
    callbacks : list
        List of callback functions to be executed during training.
    n_epochs : int
        Number of epochs for training.
    ckpt_dir:
        Directory to save checkpoints.
    ckpt_interval : int, default = 1
        Interval for saving checkpoints.
    val_ds : InMemoryDataset, default = None
        Validation dataset.
    patience : int, default = None
        Patience for early stopping.
    disable_pbar : bool, default = False
        Whether to disable progress bar for epochs..
    disable_batch_pbar : bool, default = True
        Whether to disable progress bar for batches.
    is_ensemble : bool, default = False
        Whether the model is an ensemble.
    data_parallel : bool, default = True
        Whether to use data parallelism.
    """

    log.info("Beginning Training")
    callbacks.on_train_begin()

    latest_dir = ckpt_dir / "latest"
    best_dir = ckpt_dir / "best"
    ckpt_manager = CheckpointManager()

    train_step, val_step = make_step_fns(
        loss_fn, Metrics, model=state.apply_fn, is_ensemble=is_ensemble
    )
    if train_ds.n_jit_steps > 1:
        train_step = jax.jit(functools.partial(jax.lax.scan, train_step))

    state, start_epoch = load_state(state, latest_dir)
    if start_epoch >= n_epochs:
        print(
            f"Training has already completed ({start_epoch} >= {n_epochs}). Nothing to be done"
        )
        return

    devices = len(jax.devices())
    if devices > 1 and data_parallel:
        sharding = PositionalSharding(mesh_utils.create_device_mesh((devices,)))
        state = jax.device_put(state, sharding.replicate())
    else:
        sharding = None

    train_steps_per_epoch = train_ds.steps_per_epoch()
    batch_train_ds = train_ds.shuffle_and_batch(sharding)

    if val_ds is not None:
        val_steps_per_epoch = val_ds.steps_per_epoch()
        batch_val_ds = val_ds.batch(sharding)

    best_loss = np.inf
    early_stopping_counter = 0
    epoch_loss = {}
    epoch_pbar = trange(
        start_epoch, n_epochs, desc="Epochs", ncols=100, disable=disable_pbar, leave=True
    )
    for epoch in range(start_epoch, n_epochs):
        epoch_start_time = time.time()
        callbacks.on_epoch_begin(epoch=epoch + 1)

        if ema_handler:
            ema_handler.update(state.params, epoch)

        epoch_loss.update({"train_loss": 0.0})
        train_batch_metrics = Metrics.empty()

        batch_pbar = trange(
            0,
            train_steps_per_epoch,
            desc="Batches",
            ncols=100,
            mininterval=1.0,
            disable=disable_batch_pbar,
            leave=False,
        )

        for batch_idx in range(train_steps_per_epoch):
            callbacks.on_train_batch_begin(batch=batch_idx)

            batch = next(batch_train_ds)
            (
                (state, train_batch_metrics),
                batch_loss,
            ) = train_step(
                (state, train_batch_metrics),
                batch,
            )

            epoch_loss["train_loss"] += jnp.mean(batch_loss)
            callbacks.on_train_batch_end(batch=batch_idx)
            batch_pbar.update()

        epoch_loss["train_loss"] /= train_steps_per_epoch
        epoch_loss["train_loss"] = float(epoch_loss["train_loss"])

        epoch_metrics = {
            f"train_{key}": float(val)
            for key, val in train_batch_metrics.compute().items()
        }

        if ema_handler:
            ema_handler.update(state.params, epoch)
            val_params = ema_handler.ema_params
        else:
            val_params = state.params

        if val_ds is not None:
            epoch_loss.update({"val_loss": 0.0})
            val_batch_metrics = Metrics.empty()

            batch_pbar = trange(
                0,
                val_steps_per_epoch,
                desc="Batches",
                ncols=100,
                mininterval=1.0,
                disable=disable_batch_pbar,
                leave=False,
            )
            for batch_idx in range(val_steps_per_epoch):
                batch = next(batch_val_ds)

                batch_loss, val_batch_metrics = val_step(
                    val_params, batch, val_batch_metrics
                )
                epoch_loss["val_loss"] += batch_loss
                batch_pbar.update()

            epoch_loss["val_loss"] /= val_steps_per_epoch
            epoch_loss["val_loss"] = float(epoch_loss["val_loss"])

            epoch_metrics.update(
                {
                    f"val_{key}": float(val)
                    for key, val in val_batch_metrics.compute().items()
                }
            )

        epoch_metrics.update({**epoch_loss})
        epoch_end_time = time.time()
        epoch_metrics.update({"epoch_time": epoch_end_time - epoch_start_time})

        ckpt = {"model": state, "epoch": epoch}
        if epoch % ckpt_interval == 0:
            ckpt_manager.save_checkpoint(ckpt, epoch, latest_dir)

        if epoch_metrics["val_loss"] < best_loss:
            best_loss = epoch_metrics["val_loss"]
            ckpt_manager.save_checkpoint(ckpt, epoch, best_dir)
            early_stopping_counter = 0
        else:
            early_stopping_counter += 1

        callbacks.on_epoch_end(epoch=epoch, logs=epoch_metrics)

        epoch_pbar.set_postfix(val_loss=epoch_metrics["val_loss"])
        epoch_pbar.update()

        if patience is not None and early_stopping_counter >= patience:
            log.info(
                "Early stopping patience exceeded. Stopping training after"
                f" {epoch} epochs."
            )
            break
    epoch_pbar.close()
    callbacks.on_train_end()

    train_ds.cleanup()
    if val_ds:
        val_ds.cleanup()


def calc_loss(params, inputs, labels, loss_fn, model):
    R, Z, idx, box, offsets = (
        inputs["positions"],
        inputs["numbers"],
        inputs["idx"],
        inputs["box"],
        inputs["offsets"],
    )
    predictions = model(params, R, Z, idx, box, offsets)
    loss = loss_fn(inputs, predictions, labels)
    return loss, predictions


def make_ensemble_update(update_fn: Callable) -> Callable:
    # vmap over train state
    v_update_fn = jax.vmap(update_fn, (0, None, None), (0, 0, 0))

    def ensemble_update_fn(state, inputs, labels):
        loss, predictions, state = v_update_fn(state, inputs, labels)

        mean_predictions = tree_util.tree_map(lambda x: jnp.mean(x, axis=0), predictions)
        mean_loss = jnp.mean(loss)
        # Should we add std to predictions?
        return mean_loss, mean_predictions, state

    return ensemble_update_fn


def make_ensemble_eval(update_fn: Callable) -> Callable:
    # vmap over train state
    v_update_fn = jax.vmap(update_fn, (0, None, None), (0, 0))

    def ensemble_eval_fn(state, inputs, labels):
        loss, predictions = v_update_fn(state, inputs, labels)

        mean_predictions = tree_util.tree_map(lambda x: jnp.mean(x, axis=0), predictions)
        mean_loss = jnp.mean(loss)
        return mean_loss, mean_predictions

    return ensemble_eval_fn


def make_step_fns(loss_fn, Metrics, model, is_ensemble):
    loss_calculator = partial(calc_loss, loss_fn=loss_fn, model=model)
    grad_fn = jax.value_and_grad(loss_calculator, 0, has_aux=True)

    def update_step(state, inputs, labels):
        (loss, predictions), grads = grad_fn(state.params, inputs, labels)
        state = state.apply_gradients(grads=grads)
        return loss, predictions, state

    if is_ensemble:
        update_fn = make_ensemble_update(update_step)
        eval_fn = make_ensemble_eval(loss_calculator)
    else:
        update_fn = update_step
        eval_fn = loss_calculator

    @jax.jit
    def train_step(carry, batch):
        state, batch_metrics = carry
        inputs, labels = batch
        loss, predictions, state = update_fn(state, inputs, labels)

        new_batch_metrics = Metrics.single_from_model_output(
            label=labels, prediction=predictions
        )
        batch_metrics = batch_metrics.merge(new_batch_metrics)

        new_carry = (state, batch_metrics)
        return new_carry, loss

    @jax.jit
    def val_step(params, batch, batch_metrics):
        inputs, labels = batch
        loss, predictions = eval_fn(params, inputs, labels)

        new_batch_metrics = Metrics.single_from_model_output(
            label=labels, prediction=predictions
        )
        batch_metrics = batch_metrics.merge(new_batch_metrics)
        return loss, batch_metrics

    return train_step, val_step



import jax
import jax.numpy as jnp

from apax.layers.ntk_linear import NTKLinear


def test_ntk_linear():
    key = jax.random.PRNGKey(0)

    x = jnp.linspace(-1, 1, num=20)

    linear = NTKLinear(32, dtype=jnp.float32)
    params = linear.init(key, x)
    result = linear.apply(params, x)

    assert params["params"]["w"].shape == (20, 32)
    assert params["params"]["b"].shape == (32,)

    assert result.shape == (32,)



import logging

from flax.core.frozen_dict import freeze, unfreeze
from flax.traverse_util import flatten_dict, unflatten_dict

from apax.train.checkpoints import load_params

log = logging.getLogger(__name__)


def black_list_param_transfer(source_params, target_params, param_black_list):
    source_params = unfreeze(source_params)
    target_params = unfreeze(target_params)

    flat_source = flatten_dict(source_params)
    flat_target = flatten_dict(target_params)
    for p, v in flat_source.items():
        if p[-2] not in param_black_list:
            flat_target[p] = v
            log.info("Transferring parameter: %s", p[-2])

    transfered_target = unflatten_dict(flat_target)
    transfered_target = freeze(transfered_target)
    return transfered_target


def transfer_parameters(state, ckpt_config):
    source_params = load_params(ckpt_config.base_model_checkpoint)
    log.info("Transferring parameters from %s", ckpt_config.base_model_checkpoint)
    params = black_list_param_transfer(
        source_params, state.params, ckpt_config.reset_layers
    )
    state = state.replace(params=params)
    return state



import numpy as np

from apax.utils.convert import transpose_dict_of_lists


def test_transpose_dict_of_lists():
    b = np.arange(8).reshape((4, 2))
    a = [0, 1, 2, 3]
    inputs = {
        "a": a,
        "b": b,
    }

    out = transpose_dict_of_lists(inputs)
    assert len(out) == len(a)

    for ii, entry in enumerate(out):
        assert "a" in entry.keys()
        assert "b" in entry.keys()
        assert entry["a"] == a[ii]
        assert np.all(entry["b"] == b[ii])



import collections
import csv
import logging
from pathlib import Path

import numpy as np
import tensorflow as tf
from keras.callbacks import CSVLogger, TensorBoard

from apax.config.train_config import Config

try:
    from apax.train.mlflow import MLFlowLogger
except ImportError:
    MLFlowLogger = None

log = logging.getLogger(__name__)


class CallbackCollection:
    def __init__(self, callbacks: list) -> None:
        self.callbacks = callbacks

    def on_train_begin(self, logs=None):
        for cb in self.callbacks:
            cb.on_train_begin(logs)

    def on_epoch_begin(self, epoch, logs=None):
        for cb in self.callbacks:
            cb.on_epoch_begin(epoch)

    def on_train_batch_begin(self, batch, logs=None):
        for cb in self.callbacks:
            cb.on_train_batch_begin(batch)

    def on_train_batch_end(self, batch, logs=None):
        for cb in self.callbacks:
            cb.on_train_batch_end(batch)

    def on_epoch_end(self, epoch, logs):
        for cb in self.callbacks:
            cb.on_epoch_end(epoch, logs)

    def on_train_end(self, logs=None):
        for cb in self.callbacks:
            cb.on_train_end(logs)

    def on_test_batch_end(self, batch, logs=None):
        for cb in self.callbacks:
            cb.on_test_batch_end(batch, logs)


def format_str(k):
    return f"{k:.5f}"


class CSVLoggerApax(CSVLogger):
    def __init__(self, filename, separator=",", append=False):
        super().__init__(filename, separator=separator, append=append)

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}

        def handle_value(k):
            is_zero_dim_ndarray = isinstance(k, np.ndarray) and k.ndim == 0
            if isinstance(k, str):
                return k
            elif isinstance(k, collections.abc.Iterable) and not is_zero_dim_ndarray:
                return f'"[{", ".join(map(format_str, k))}]"'
            else:
                return format_str(k)

        if self.keys is None:
            self.keys = sorted(logs.keys())
            # When validation_freq > 1, `val_` keys are not in first epoch logs
            # Add the `val_` keys so that its part of the fieldnames of writer.
            val_keys_found = False
            for key in self.keys:
                if key.startswith("val_"):
                    val_keys_found = True
                    break
            if not val_keys_found:
                self.keys.extend(["val_" + k for k in self.keys])

        if not self.writer:

            class CustomDialect(csv.excel):
                delimiter = self.sep

            fieldnames = ["epoch"] + self.keys

            self.writer = csv.DictWriter(
                self.csv_file, fieldnames=fieldnames, dialect=CustomDialect
            )
            if self.append_header:
                self.writer.writeheader()

        row_dict = collections.OrderedDict({"epoch": epoch})
        row_dict.update((key, handle_value(logs.get(key, "NA"))) for key in self.keys)
        self.writer.writerow(row_dict)
        self.csv_file.flush()

    def on_test_batch_end(self, batch, logs=None):
        logs = logs or {}

        def handle_value(k):
            is_zero_dim_ndarray = isinstance(k, np.ndarray) and k.ndim == 0
            if isinstance(k, str):
                return k
            elif isinstance(k, collections.abc.Iterable) and not is_zero_dim_ndarray:
                return f'"[{", ".join(map(format_str, k))}]"'
            else:
                return format_str(k)

        if self.keys is None:
            self.keys = sorted(logs.keys())

        if not self.writer:

            class CustomDialect(csv.excel):
                delimiter = self.sep

            fieldnames = ["batch"] + self.keys

            self.writer = csv.DictWriter(
                self.csv_file, fieldnames=fieldnames, dialect=CustomDialect
            )
            if self.append_header:
                self.writer.writeheader()

        row_dict = collections.OrderedDict({"batch": batch})
        row_dict.update((key, handle_value(logs.get(key, "NA"))) for key in self.keys)
        self.writer.writerow(row_dict)
        self.csv_file.flush()


def initialize_callbacks(config: Config, model_version_path: Path):
    callback_configs = config.callbacks
    log.info("Initializing Callbacks")

    dummy_model = tf.keras.Model()
    dummy_model.compile(loss="mse", optimizer="adam")
    callback_dict = {
        "csv": {
            "class": CSVLoggerApax,
            "log_path": model_version_path / "log.csv",
            "path_arg_name": "filename",
            "kwargs": {"append": True},
            "model": dummy_model,
        },
        "tensorboard": {
            "class": TensorBoard,
            "log_path": model_version_path,
            "path_arg_name": "log_dir",
            "kwargs": {},
            "model": dummy_model,
            "write_graph": False,
        },
        "mlflow": {
            "class": MLFlowLogger,
            "log_path": model_version_path,
            "path_arg_name": "log_dir",
            "kwargs": {"run_name": config.data.experiment},
        },
    }

    callbacks = []
    for callback_config in callback_configs:
        if callback_config.name == "mlflow":
            callback = MLFlowLogger(
                experiment=callback_config.experiment, run_name=config.data.experiment
            )
        else:
            callback_info = callback_dict[callback_config.name]

            path_arg_name = callback_info["path_arg_name"]
            path = {path_arg_name: callback_info["log_path"]}

            kwargs = callback_info["kwargs"]
            callback = callback_info["class"](**path, **kwargs)
            callback.set_model(callback_info["model"])
        callbacks.append(callback)

    return CallbackCollection(callbacks)



def mask_by_atom(arr, Z):
    mask = (Z != 0).astype(arr.dtype)
    axes_to_add = len(arr.shape) - 1
    for _ in range(axes_to_add):
        mask = mask[..., None]
    masked_arr = arr * mask
    return masked_arr


def mask_by_neighbor(arr, idx):
    mask = ((idx[0] - idx[1]) != 0).astype(arr.dtype)
    if len(arr.shape) == 2:
        mask = mask[..., None]
    elif len(arr.shape) == 4:
        mask = mask[:, None, None, None]
    return arr * mask



import pathlib

import jax

from apax.config.common import parse_config
from apax.utils.data import make_minimal_input

TEST_PATH = pathlib.Path(__file__).parent.resolve()


def test_builder_feature_model():
    R, Z, idx, box, offsets = make_minimal_input()

    config = parse_config(TEST_PATH / "train.yaml")

    Builder = config.model.get_builder()
    builder = Builder(config.model.model_dump())

    model = builder.build_energy_model()
    key = jax.random.PRNGKey(0)
    params = model.init(key, R, Z, idx, box, offsets)

    model = builder.build_feature_model(only_use_n_layers=0, init_box=box)
    out = model.apply(params, R, Z, idx, box, offsets)
    assert out.shape == (360,)

    model = builder.build_feature_model(only_use_n_layers=1, init_box=box)
    out = model.apply(params, R, Z, idx, box, offsets)
    assert out.shape == (128,)

    model = builder.build_feature_model(only_use_n_layers=2, init_box=box)
    out = model.apply(params, R, Z, idx, box, offsets)
    assert out.shape == (64,)

    model = builder.build_feature_model(only_use_n_layers=3, init_box=box)
    out = model.apply(params, R, Z, idx, box, offsets)
    assert out.shape == (32,)



from typing import Any

import einops
import flax.linen as nn
import jax.numpy as jnp
from jax import vmap

from apax.layers.descriptor.basis_functions import RadialFunction
from apax.layers.descriptor.moments import geometric_moments
from apax.layers.descriptor.triangular_indices import tril_2d_indices, tril_3d_indices
from apax.layers.masking import mask_by_neighbor
from apax.utils.convert import str_to_dtype
from apax.utils.jax_md_reduced import space


class GaussianMomentDescriptor(nn.Module):
    radial_fn: nn.Module = RadialFunction()
    n_contr: int = 8
    dtype: Any = jnp.float32
    apply_mask: bool = True

    def setup(self):
        self.r_max = self.radial_fn.r_max
        self.n_radial = self.radial_fn._n_radial

        self.distance = vmap(space.distance, 0, 0)

        self.triang_idxs_2d = tril_2d_indices(self.n_radial)
        self.triang_idxs_3d = tril_3d_indices(self.n_radial)

    def __call__(self, dr_vec, Z, neighbor_idxs):
        dtype = str_to_dtype(self.dtype)
        dr_vec = dr_vec.astype(dtype)
        # Z shape n_atoms
        n_atoms = Z.shape[0]

        idx_i, idx_j = neighbor_idxs[0], neighbor_idxs[1]

        # shape: neighbors
        Z_i, Z_j = Z[idx_i, ...], Z[idx_j, ...]

        # dr shape: neighbors
        dr = self.distance(dr_vec)

        # TODO: maybe try jnp where
        dr_repeated = einops.repeat(dr + 1e-5, "neighbors -> neighbors 1")
        # normalized distance vectors, shape neighbors x 3
        dn = dr_vec / dr_repeated

        radial_function = self.radial_fn(dr, Z_i, Z_j)
        if self.apply_mask:
            radial_function = mask_by_neighbor(radial_function, neighbor_idxs)

        moments = geometric_moments(radial_function, dn, idx_j, n_atoms)

        contr_0 = moments[0]
        contr_1 = jnp.einsum("ari, asi -> ars", moments[1], moments[1])  # noqa: E501
        contr_2 = jnp.einsum("arij, asij -> ars", moments[2], moments[2])  # noqa: E501
        contr_3 = jnp.einsum("arijk, asijk -> ars", moments[3], moments[3])  # noqa: E501
        contr_4 = jnp.einsum(
            "arij, asik, atjk -> arst", moments[2], moments[2], moments[2]
        )  # noqa: E501
        contr_5 = jnp.einsum("ari, asj, atij -> arst", moments[1], moments[1], moments[2])  # noqa: E501
        contr_6 = jnp.einsum(
            "arijk, asijl, atkl -> arst", moments[3], moments[3], moments[2]
        )  # noqa: E501
        contr_7 = jnp.einsum(
            "arijk, asij, atk -> arst", moments[3], moments[2], moments[1]
        )  # noqa: E501

        tril_2_i, tril_2_j = self.triang_idxs_2d[:, 0], self.triang_idxs_2d[:, 1]
        tril_3_i, tril_3_j, tril_3_k = (
            self.triang_idxs_3d[:, 0],
            self.triang_idxs_3d[:, 1],
            self.triang_idxs_3d[:, 2],
        )

        contr_1 = contr_1[:, tril_2_i, tril_2_j]
        contr_2 = contr_2[:, tril_2_i, tril_2_j]
        contr_3 = contr_3[:, tril_2_i, tril_2_j]
        contr_4 = contr_4[:, tril_3_i, tril_3_j, tril_3_k]
        contr_5 = contr_5[:, tril_2_i, tril_2_j]
        contr_6 = contr_6[:, tril_2_i, tril_2_j]

        contr_5 = jnp.reshape(contr_5, [n_atoms, -1])
        contr_6 = jnp.reshape(contr_6, [n_atoms, -1])
        contr_7 = jnp.reshape(contr_7, [n_atoms, -1])

        gaussian_moments = [
            contr_0,
            contr_1,
            contr_2,
            contr_3,
            contr_4,
            contr_5,
            contr_6,
            contr_7,
        ]

        # gaussian_moments shape: n_atoms x n_features
        gaussian_moments = jnp.concatenate(gaussian_moments[: self.n_contr], axis=-1)
        assert gaussian_moments.dtype == dtype
        return gaussian_moments



import pathlib
import uuid

import numpy as np
import pytest

from tests.conftest import load_config_and_run_training

TEST_PATH = pathlib.Path(__file__).parent.resolve()


def load_csv(filename):
    data = np.loadtxt(filename, delimiter=",", skiprows=1)  # Skip the header row

    with open(filename, "r") as file:
        header = file.readline().strip().split(",")

    data_dict = {header[i]: data[:, i].tolist() for i in range(len(header))}

    return data_dict


@pytest.mark.slow
def test_regression_model_training(get_md22_stachyose, get_tmp_path):
    config_path = TEST_PATH / "apax_config.yaml"
    working_dir = get_tmp_path / str(uuid.uuid4())
    file_path = get_md22_stachyose

    data_config_mods = {
        "data": {
            "directory": working_dir.as_posix(),
            "data_path": file_path.as_posix(),
            "energy_unit": "kcal/mol",
        }
    }

    load_config_and_run_training(config_path, data_config_mods)

    current_metrics = load_csv(working_dir / "test/log.csv")

    comparison_metrics = {
        "val_energy_mae": 0.075,
        "val_forces_mae": 0.045,
        "val_forces_mse": 0.004,
        "val_loss": 0.045,
    }

    for key in comparison_metrics.keys():
        print((np.array(current_metrics[key])[-1]))

    for key in comparison_metrics.keys():
        assert abs((np.array(current_metrics[key])[-1] - comparison_metrics[key])) < 0.01
